'''
Define utilities for building phases.

Classes
-------
PhaseBuilderBase : the interface for a phase builder
'''
from abc import ABC
from collections import namedtuple

from aviary.mission.initial_guess_builders import InitialGuess
from aviary.variable_info.variables import Dynamic

import dymos as dm
import openmdao.api as om

from aviary.mission.flops_based.ode.mission_ODE import MissionODE
from aviary.utils.aviary_values import AviaryValues, get_keys
from aviary.variable_info.variable_meta_data import _MetaData

_require_new_meta_data_class_attr_ = \
    namedtuple('_require_new_meta_data_class_attr_', ())


_require_new_initial_guesses_meta_data_class_attr_ = \
    namedtuple('_require_new_initial_guesses_meta_data_class_attr_', ())


class PhaseBuilderBase(ABC):
    '''
    Define the interface for a phase builder.

    Attributes
    ----------
    name : str ('<unknown phase>')
        object label

    core_subsystems : (None)
        list of SubsystemBuilderBase objects that will be added to the phase ODE

    user_options : AviaryValues (<empty>)
        state/path constraint values and flags

    initial_guesses : AviaryValues (<empty>)
        state/path beginning values to be set on the problem

    ode_class : type (None)
        advanced: the type of system defining the ODE

    transcription : "Dymos transcription object" (None)
        advanced: an object providing the transcription technique of the
        optimal control problem

    subsystem_options : dict (None)
        dictionary of parameters to be passed to the subsystem builders

    default_name : str
        class attribute: derived type customization point; the default value
        for name

    default_ode_class : type
        class attribute: derived type customization point; the default value
        for ode_class used by build_phase

    is_analytic_phase : bool (False)
        class attribute: derived type customization point; if True, build_phase
        will return an AnalyticPhase instead of a Phase

    num_nodes : int (5)
        class attribute: derived type customization point; the default value
        for num_nodes used by build_phase, only for AnalyticPhases

    Methods
    -------
    build_phase
    make_default_transcription
    validate_options
    assign_default_options
    '''
    __slots__ = (
        'name',  'core_subsystems', 'subsystem_options', 'user_options',
        'initial_guesses', 'ode_class', 'transcription',
        'is_analytic_phase', 'num_nodes', 'external_subsystems', 'meta_data',
    )

    # region : derived type customization points
    _meta_data_ = _require_new_meta_data_class_attr_()

    _initial_guesses_meta_data_ = _require_new_initial_guesses_meta_data_class_attr_()

    default_name = '<unknown phase>'

    default_ode_class = MissionODE

    default_meta_data = _MetaData
    # endregion : derived type customization points

    def __init__(
        self, name=None, core_subsystems=None, user_options=None, initial_guesses=None,
        ode_class=None, transcription=None, subsystem_options=None, is_analytic_phase=False, num_nodes=5, external_subsystems=None, meta_data=None,
    ):
        if name is None:
            name = self.default_name

        self.name = name

        if core_subsystems is None:
            core_subsystems = []

        self.core_subsystems = core_subsystems

        if subsystem_options is None:
            subsystem_options = {}

        self.subsystem_options = subsystem_options

        self.user_options = user_options
        self.validate_options()
        self.assign_default_options()

        if initial_guesses is None:
            initial_guesses = AviaryValues()

        self.initial_guesses = initial_guesses
        self.validate_initial_guesses()

        self.ode_class = ode_class
        self.transcription = transcription
        self.is_analytic_phase = is_analytic_phase
        self.num_nodes = num_nodes

        if external_subsystems is None:
            external_subsystems = []

        self.external_subsystems = external_subsystems

        if meta_data is None:
            meta_data = self.default_meta_data

        self.meta_data = meta_data

    def build_phase(self, aviary_options=None):
        '''
        Return a new phase object for analysis using these constraints.

        If ode_class is None, default_ode_class is used.

        If transcription is None, the return value from calling
        make_default_transcription is used.

        Parameters
        ----------
        aviary_options : AviaryValues (emtpy)
            collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        '''
        ode_class = self.ode_class

        if ode_class is None:
            ode_class = self.default_ode_class

        transcription = self.transcription

        if transcription is None and not self.is_analytic_phase:
            transcription = self.make_default_transcription()

        if aviary_options is None:
            aviary_options = AviaryValues()

        kwargs = self._extra_ode_init_kwargs()

        kwargs = {
            'aviary_options': aviary_options,
            **kwargs
        }

        subsystem_options = self.subsystem_options

        if subsystem_options is not None:
            kwargs['subsystem_options'] = subsystem_options

        kwargs['core_subsystems'] = self.core_subsystems

        if self.is_analytic_phase:
            phase = dm.AnalyticPhase(
                ode_class=ode_class,
                ode_init_kwargs=kwargs,
                num_nodes=self.num_nodes,
            )
        else:
            phase = dm.Phase(
                ode_class=ode_class, transcription=transcription,
                ode_init_kwargs=kwargs
            )

        # overrides should add state, controls, etc.
        return phase

    def make_default_transcription(self):
        '''
        Return a transcription object to be used by default in build_phase.
        '''
        user_options = self.user_options

        num_segments, _ = user_options.get_item('num_segments')
        order, _ = user_options.get_item('order')

        transcription = dm.Radau(
            num_segments=num_segments, order=order, compressed=True)

        return transcription

    def validate_options(self):
        '''
        Raise TypeError if an unsupported option is found.

        Users can call this method when updating options after initialization.
        '''
        user_options = self.user_options

        if not user_options:
            return  # acceptable

        meta_data = self._meta_data_

        for key in get_keys(user_options):
            if key not in meta_data:
                raise TypeError(
                    f'{self.__class__.__name__}: {self.name}:'
                    f' unsupported option: {key}'
                )

    def assign_default_options(self):
        '''
        Update missing options with default values.

        If user_options is None, start with an empty AviaryValues.

        Users can call this method when replacing the user_options member after
        initialization.
        '''
        user_options = self.user_options

        if user_options is None:
            user_options = self.user_options = AviaryValues()

        meta_data = self._meta_data_

        for key in meta_data:
            if key not in user_options:
                item = meta_data[key]

                val = item['val']
                units = item['units']

                user_options.set_val(key, val, units)

    def validate_initial_guesses(self):
        '''
        Raise TypeError if an unsupported initial guess is found.

        Users can call this method when updating initial guesses after initialization.
        '''
        initial_guesses = self.initial_guesses

        if not initial_guesses:
            return  # acceptable

        meta_data = self._initial_guesses_meta_data_

        for key in get_keys(initial_guesses):
            if key not in meta_data:
                raise TypeError(
                    f'{self.__class__.__name__}: {self.name}:'
                    f' unsupported initial guess: {key}'
                )

    def apply_initial_guesses(
        self, prob: om.Problem, traj_name, phase: dm.Phase
    ):
        '''
        Apply any stored initial guesses; return a list of guesses not applied.
        '''
        not_applied = []

        phase_name = self.name
        meta_data = self._initial_guesses_meta_data_
        initial_guesses: AviaryValues = self.initial_guesses

        for key in meta_data:
            if key in initial_guesses:
                apply_initial_guess = meta_data[key]['apply_initial_guess']
                val, units = initial_guesses.get_item(key)

                apply_initial_guess(prob, traj_name, phase, phase_name, val, units)

            else:
                not_applied.append(key)

        return not_applied

    def _extra_ode_init_kwargs(self):
        """
        Return extra kwargs required for initializing the ODE.
        """
        return {}

    def to_phase_info(self):
        '''
        Return the stored settings as phase info.

        Returns
        -------
        tuple
            name : str
                object label
            phase_info : dict
                stored settings
        '''
        subsystem_options = self.subsystem_options  # TODO: aero info?
        user_options = dict(self.user_options)
        initial_guesses = dict(self.initial_guesses)

        # TODO some of these may be purely programming API hooks, rather than for use
        # with phase info
        # - ode_class
        # - transcription
        # - external_subsystems
        # - meta_data

        phase_info = dict(
            subsystem_options=subsystem_options, user_options=user_options,
            initial_guesses=initial_guesses)

        return (self.name, phase_info)

    @classmethod
    def from_phase_info(cls, name, phase_info: dict, core_subsystems=None, meta_data=None, transcription=None):
        '''
        Return a new phase builder based on the specified phase info.

        Note, calling code is responsible for matching phase info to the correct phase
        builder type, or the behavior is undefined.

        Parameters
        ----------
        name : str
            object label
        phase_info : dict
            stored settings
        '''
        # loop over user_options dict entries
        # if the value is not a tuple, wrap it in a tuple with the second entry of 'unitless'
        for key, value in phase_info['user_options'].items():
            if not isinstance(value, tuple):
                phase_info['user_options'][key] = (value, 'unitless')

        subsystem_options = phase_info.get(
            'subsystem_options', {})  # TODO: aero info?
        user_options = AviaryValues(phase_info.get('user_options', ()))
        initial_guesses = AviaryValues(phase_info.get('initial_guesses', ()))
        external_subsystems = phase_info.get('external_subsystems', [])
        # TODO core subsystems in phase info?

        # TODO some of these may be purely programming API hooks, rather than for use
        # with phase info
        # - ode_class
        # - transcription
        # - external_subsystems
        # - meta_data

        phase_builder = cls(
            name, subsystem_options=subsystem_options, user_options=user_options,
            initial_guesses=initial_guesses, meta_data=meta_data,
            core_subsystems=core_subsystems, external_subsystems=external_subsystems, transcription=transcription)

        return phase_builder

    @classmethod
    def _add_meta_data(cls, name, *, val, units='unitless', desc=None):
        '''
        Update supported options with a new item.

        Raises
        ------
        ValueError
            if a repeat option is found
        '''
        meta_data = cls._meta_data_

        if name in meta_data:
            raise ValueError(
                f'{cls.__name__}": meta data: repeat option: {name}'
            )

        meta_data[name] = dict(val=val, units=units, desc=desc)

    @classmethod
    def _add_initial_guess_meta_data(cls, initial_guess: InitialGuess, desc=None):
        '''
        Update supported initial guesses with a new item.

        Raises
        ------
        ValueError
            if a repeat initial guess is found
        '''
        meta_data = cls._initial_guesses_meta_data_
        name = initial_guess.key

        if name in meta_data:
            raise ValueError(
                f'{cls.__name__}": meta data: repeat initial guess: {name}'
            )

        meta_data[name] = dict(
            apply_initial_guess=initial_guess.apply_initial_guess, desc=desc)

    def _add_user_defined_constraints(self, phase, constraints):
        # Add each constraint and its corresponding arguments to the phase
        for constraint_name, kwargs in constraints.items():
            if kwargs['type'] == 'boundary':
                kwargs.pop('type')
                phase.add_boundary_constraint(constraint_name, **kwargs)
            elif kwargs['type'] == 'path':
                kwargs.pop('type')
                phase.add_path_constraint(constraint_name, **kwargs)

    def set_time_options(self, user_options, targets=[]):
        fix_initial = user_options.get_val('fix_initial')
        duration_bounds = user_options.get_val('duration_bounds', units='s')
        duration_ref = user_options.get_val('duration_ref', units='s')

        self.phase.set_time_options(
            fix_initial=fix_initial,
            duration_bounds=duration_bounds,
            units="s",
            targets=targets,
            duration_ref=duration_ref,
        )

    def add_velocity_state(self, user_options):
        velocity_lower = user_options.get_val('velocity_lower', units='kn')
        velocity_upper = user_options.get_val('velocity_upper', units='kn')
        velocity_ref = user_options.get_val('velocity_ref', units='kn')
        velocity_ref0 = user_options.get_val('velocity_ref0', units='kn')
        velocity_defect_ref = user_options.get_val('velocity_defect_ref', units='kn')
        self.phase.add_state(
            Dynamic.Mission.VELOCITY,
            fix_initial=user_options.get_val('fix_initial'),
            fix_final=False,
            lower=velocity_lower,
            upper=velocity_upper,
            units="kn",
            rate_source=Dynamic.Mission.VELOCITY_RATE,
            targets=Dynamic.Mission.VELOCITY,
            ref=velocity_ref,
            ref0=velocity_ref0,
            defect_ref=velocity_defect_ref,
        )

    def add_mass_state(self, user_options):
        mass_lower = user_options.get_val('mass_lower', units='lbm')
        mass_upper = user_options.get_val('mass_upper', units='lbm')
        mass_ref = user_options.get_val('mass_ref', units='lbm')
        mass_ref0 = user_options.get_val('mass_ref0', units='lbm')
        mass_defect_ref = user_options.get_val('mass_defect_ref', units='lbm')
        self.phase.add_state(
            Dynamic.Mission.MASS,
            fix_initial=user_options.get_val('fix_initial'),
            fix_final=False,
            lower=mass_lower,
            upper=mass_upper,
            units="lbm",
            rate_source=Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            targets=Dynamic.Mission.MASS,
            ref=mass_ref,
            ref0=mass_ref0,
            defect_ref=mass_defect_ref,
        )

    def add_distance_state(self, user_options, units='NM'):
        distance_lower = user_options.get_val('distance_lower', units=units)
        distance_upper = user_options.get_val('distance_upper', units=units)
        distance_ref = user_options.get_val('distance_ref', units=units)
        distance_ref0 = user_options.get_val('distance_ref0', units=units)
        distance_defect_ref = user_options.get_val('distance_defect_ref', units=units)
        self.phase.add_state(
            Dynamic.Mission.DISTANCE,
            fix_initial=user_options.get_val('fix_initial'),
            fix_final=False,
            lower=distance_lower,
            upper=distance_upper,
            units=units,
            rate_source=Dynamic.Mission.DISTANCE_RATE,
            ref=distance_ref,
            ref0=distance_ref0,
            defect_ref=distance_defect_ref,
        )

    def add_flight_path_angle_state(self, user_options):
        angle_lower = user_options.get_val('angle_lower', units='rad')
        angle_upper = user_options.get_val('angle_upper', units='rad')
        angle_ref = user_options.get_val('angle_ref', units='rad')
        angle_ref0 = user_options.get_val('angle_ref0', units='rad')
        angle_defect_ref = user_options.get_val('angle_defect_ref', units='rad')
        self.phase.add_state(
            Dynamic.Mission.FLIGHT_PATH_ANGLE,
            fix_initial=True,
            fix_final=False,
            lower=angle_lower,
            upper=angle_upper,
            units="rad",
            rate_source=Dynamic.Mission.FLIGHT_PATH_ANGLE_RATE,
            ref=angle_ref,
            defect_ref=angle_defect_ref,
            ref0=angle_ref0,
        )

    def add_altitude_state(self, user_options, units='ft'):
        alt_lower = user_options.get_val('alt_lower', units=units)
        alt_upper = user_options.get_val('alt_upper', units=units)
        alt_ref = user_options.get_val('alt_ref', units=units)
        alt_ref0 = user_options.get_val('alt_ref0', units=units)
        alt_defect_ref = user_options.get_val('alt_defect_ref', units=units)
        self.phase.add_state(
            Dynamic.Mission.ALTITUDE,
            fix_initial=True,
            fix_final=False,
            lower=alt_lower,
            upper=alt_upper,
            units=units,
            rate_source=Dynamic.Mission.ALTITUDE_RATE,
            ref=alt_ref,
            defect_ref=alt_defect_ref,
            ref0=alt_ref0,
        )

    def add_altitude_constraint(self, user_options):
        final_altitude = user_options.get_val('final_altitude', units='ft')
        alt_constraint_ref = user_options.get_val('alt_constraint_ref', units='ft')
        self.phase.add_boundary_constraint(
            Dynamic.Mission.ALTITUDE,
            loc="final",
            equals=final_altitude,
            units="ft",
            ref=alt_constraint_ref,
        )


_registered_phase_builder_types = []


def register(phase_builder_t=None, *, check_repeats=True):
    '''
    Register a new phase builder type.

    Note, this function qualifies as a class decorator for ease of use.

    Returns
    -------
    '''
    if phase_builder_t is None:
        def decorator(phase_builder_t):
            return register(phase_builder_t, check_repeats=check_repeats)

        return decorator

    if check_repeats and (phase_builder_t in _registered_phase_builder_types):
        raise ValueError('repeated phase builder type')

    _registered_phase_builder_types.append(phase_builder_t)

    return phase_builder_t


def phase_info_to_builder(name: str, phase_info: dict) -> PhaseBuilderBase:
    '''
    Return a new phase builder based on the specified phase info.

    Note, the type of phase builder will be determined by calling
    phase_builder_t.from_phase_info() for each registered type in order of registration;
    the first result that is not None will be returned. If a supported phase builder type
    cannot be determined, raise ValueError.

    Raises
    ------
    ValueError
        if a supported phase builder type cannot be determined
    '''
    phase_builder_t: PhaseBuilderBase = None

    for phase_builder_t in _registered_phase_builder_types:
        builder = phase_builder_t.from_phase_info(name, phase_info)

        if builder is not None:
            return builder

    raise ValueError(f'unsupported phase info: {name}')


if __name__ == '__main__':
    help(PhaseBuilderBase)

    try:
        PhaseBuilderBase._add_meta_data('test', val=0, units=None, desc='test')

    except Exception as error:
        print(error)




import numpy as np
from openmdao.utils.assert_utils import assert_near_equal

from aviary.utils.test_utils.assert_utils import warn_timeseries_near_equal


def compare_against_expected_values(prob, expected_dict):

    expected_times = expected_dict['times']
    expected_altitudes = expected_dict['altitudes']
    expected_masses = expected_dict['masses']
    expected_ranges = expected_dict['ranges']
    expected_velocities = expected_dict['velocities']

    times = []
    altitudes = []
    masses = []
    ranges = []
    velocities = []

    for idx, phase in enumerate(['climb', 'cruise', 'descent']):

        times.extend(prob.get_val(f'traj.{phase}.timeseries.time', units='s'))
        try:
            altitudes.extend(prob.get_val(
                f'traj.{phase}.timeseries.altitude', units='m'))
        except KeyError:
            altitudes.extend(prob.get_val(
                f'traj.{phase}.timeseries.altitude', units='m'))
        velocities.extend(prob.get_val(
            f'traj.{phase}.timeseries.velocity', units='m/s'))
        masses.extend(
            prob.get_val(f'traj.{phase}.timeseries.mass', units='kg'))
        ranges.extend(
            prob.get_val(f'traj.{phase}.timeseries.distance', units='m'))

    times = np.array(times)
    altitudes = np.array(altitudes)
    masses = np.array(masses)
    ranges = np.array(ranges)
    velocities = np.array(velocities)

    # Check Objective and other key variables to a reasonably tight tolerance.

    rtol = 2.e-2

    # Mass at the end of Descent
    assert_near_equal(masses[-1], expected_masses[-1], tolerance=rtol)

    # Range at the end of Descent
    assert_near_equal(ranges[-1], expected_ranges[-1], tolerance=rtol)

    # Flight time
    assert_near_equal(times[-1], expected_times[-1], tolerance=rtol)

    # Check mission values.

    # NOTE rtol = 0.05 = 5% different from truth (first timeseries)
    #      atol = 2 = no more than +/-2 meter/second/kg difference between values
    #      atol_altitude - 30 ft.
    rtol = .05
    atol = 2.0
    atol_altitude = 30.0

    # FLIGHT PATH
    warn_timeseries_near_equal(
        times, altitudes, expected_times,
        expected_altitudes, abs_tolerance=atol_altitude, rel_tolerance=rtol)
    warn_timeseries_near_equal(
        times, masses, expected_times,
        expected_masses, abs_tolerance=atol, rel_tolerance=rtol)
    warn_timeseries_near_equal(
        times, ranges, expected_times,
        expected_ranges, abs_tolerance=atol, rel_tolerance=rtol)
    warn_timeseries_near_equal(
        times, velocities, expected_times,
        expected_velocities, abs_tolerance=atol, rel_tolerance=rtol)



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class PropulsionMission(om.Group):
    '''
    Group that tracks all engine models used during mission analysis. Accounts for
    number of engines for each type and returns aircraft-total dynamic values such
    as net thrust and fuel flow rate.
    '''

    def initialize(self):
        self.options.declare(
            'num_nodes',
            types=int,
            lower=0
        )

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        nn = self.options['num_nodes']
        options: AviaryValues = self.options['aviary_options']
        engine_models = options.get_val('engine_models')
        count = len(engine_models)

        # TODO what if "engine" is not an EngineModel object? Type is never checked/enforced
        for (i, engine) in enumerate(engine_models):
            self.add_subsystem(
                engine.name,
                subsys=engine.build_mission(
                    num_nodes=nn, aviary_inputs=options),
                promotes_inputs=['*'])

            if count > 1:
                # split vectorized throttles and connect to the correct engine model
                self.promotes(
                    engine.name,
                    inputs=[Dynamic.Mission.THROTTLE],
                    src_indices=om.slicer[:, i])
                # TODO if only some engine use hybrid throttle, source vector will have an
                #      index for that engine that is unused, will this confuse optimizer?
                if engine.use_hybrid_throttle:
                    self.promotes(
                        engine.name,
                        inputs=[Dynamic.Mission.HYBRID_THROTTLE],
                        src_indices=om.slicer[:, i])
            else:
                self.promotes(
                    engine.name,
                    inputs=[Dynamic.Mission.THROTTLE])
                if engine.use_hybrid_throttle:
                    self.promotes(
                        engine.name,
                        inputs=[Dynamic.Mission.HYBRID_THROTTLE])

        # TODO this might be able to be automated using propulsion Enums
        # mux component to vectorize individual outputs into 2d arrays
        perf_mux = om.MuxComp(vec_size=count)
        # add each engine data variable to mux component
        perf_mux.add_var(
            Dynamic.Mission.THRUST,
            shape=(nn,),
            axis=1,
            units='lbf')
        perf_mux.add_var(
            Dynamic.Mission.THRUST_MAX,
            shape=(nn,),
            axis=1,
            units='lbf')
        perf_mux.add_var(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
            shape=(nn,),
            axis=1,
            units='lbm/h')
        perf_mux.add_var(
            Dynamic.Mission.ELECTRIC_POWER,
            shape=(nn,),
            axis=1,
            units='kW')
        perf_mux.add_var(
            Dynamic.Mission.NOX_RATE,
            shape=(nn,),
            axis=1,
            units='lb/h')
        perf_mux.add_var(
            Dynamic.Mission.TEMPERATURE_ENGINE_T4,
            shape=(nn,),
            axis=1,
            units='degR'
        )
        # perf_mux.add_var(
        #     'exit_area_unscaled',
        #     shape=(nn,),
        #     axis=1,
        #     units='ft**2')

        self.add_subsystem('vectorize_performance',
                           subsys=perf_mux,
                           promotes_outputs=['*'])

        # connect engine outputs to mux component inputs
        for (i, engine) in enumerate(engine_models):
            self.connect(engine.name + '.' + Dynamic.Mission.THRUST,
                         'vectorize_performance.' + Dynamic.Mission.THRUST + '_' + str(i))
            self.connect(engine.name + '.' + Dynamic.Mission.THRUST_MAX,
                         'vectorize_performance.' + Dynamic.Mission.THRUST_MAX + '_' + str(i))
            self.connect(engine.name + '.' + Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
                         'vectorize_performance.' + Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE + '_' + str(i))
            self.connect(engine.name + '.' + Dynamic.Mission.ELECTRIC_POWER,
                         'vectorize_performance.' + Dynamic.Mission.ELECTRIC_POWER + '_' + str(i))
            self.connect(engine.name + '.' + Dynamic.Mission.NOX_RATE,
                         'vectorize_performance.' + Dynamic.Mission.NOX_RATE + '_' + str(i))

        self.add_subsystem(
            'propulsion_sum',
            subsys=PropulsionSum(
                num_nodes=nn,
                aviary_options=options),
            promotes_inputs=['*'],
            promotes_outputs=['*']
        )


class PropulsionSum(om.ExplicitComponent):
    '''
    Calculates propulsion system level sums of individual engine performance parameters.
    '''

    def initialize(self):
        self.options.declare(
            'num_nodes',
            types=int,
            lower=0
        )

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        nn = self.options['num_nodes']
        count = len(self.options['aviary_options'].get_val('engine_models'))

        self.add_input(Dynamic.Mission.THRUST, val=np.zeros((nn, count)), units='lbf')
        self.add_input(Dynamic.Mission.THRUST_MAX,
                       val=np.zeros((nn, count)), units='lbf')
        self.add_input(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE, val=np.zeros((nn, count)),
                       units='lbm/h')
        self.add_input(Dynamic.Mission.ELECTRIC_POWER,
                       val=np.zeros((nn, count)), units='kW')
        self.add_input(Dynamic.Mission.NOX_RATE,
                       val=np.zeros((nn, count)), units='lbm/h')

        self.add_output(Dynamic.Mission.THRUST_TOTAL, val=np.zeros(nn), units='lbf')
        self.add_output(Dynamic.Mission.THRUST_MAX_TOTAL, val=np.zeros(nn), units='lbf')
        self.add_output(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, val=np.zeros(nn),
                        units='lbm/h')
        self.add_output(Dynamic.Mission.ELECTRIC_POWER_TOTAL,
                        val=np.zeros(nn), units='kW')
        self.add_output(Dynamic.Mission.NOX_RATE_TOTAL, val=np.zeros(nn), units='lbm/h')

    def setup_partials(self):
        nn = self.options['num_nodes']
        num_engines = self.options['aviary_options'].get_val(Aircraft.Engine.NUM_ENGINES)
        count = len(num_engines)
        deriv = np.tile(num_engines, nn)

        r = np.repeat(np.arange(nn, dtype=int), count)
        c = np.arange(nn * count, dtype=int)

        self.declare_partials(
            Dynamic.Mission.THRUST_TOTAL,
            Dynamic.Mission.THRUST,
            val=deriv, rows=r, cols=c)
        self.declare_partials(
            Dynamic.Mission.THRUST_MAX_TOTAL,
            Dynamic.Mission.THRUST_MAX, val=deriv, rows=r, cols=c)
        self.declare_partials(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE, val=deriv, rows=r, cols=c)
        self.declare_partials(
            Dynamic.Mission.ELECTRIC_POWER_TOTAL,
            Dynamic.Mission.ELECTRIC_POWER, val=deriv, rows=r, cols=c)
        self.declare_partials(
            Dynamic.Mission.NOX_RATE_TOTAL,
            Dynamic.Mission.NOX_RATE,
            val=deriv, rows=r, cols=c)

    def compute(self, inputs, outputs):
        num_engines = self.options['aviary_options'].get_val(Aircraft.Engine.NUM_ENGINES)

        thrust = inputs[Dynamic.Mission.THRUST]
        thrust_max = inputs[Dynamic.Mission.THRUST_MAX]
        fuel_flow = inputs[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE]
        electric = inputs[Dynamic.Mission.ELECTRIC_POWER]
        nox = inputs[Dynamic.Mission.NOX_RATE]

        outputs[Dynamic.Mission.THRUST_TOTAL] = np.dot(thrust, num_engines)
        outputs[Dynamic.Mission.THRUST_MAX_TOTAL] = np.dot(thrust_max, num_engines)
        outputs[Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL] = np.dot(
            fuel_flow, num_engines)
        outputs[Dynamic.Mission.ELECTRIC_POWER_TOTAL] = np.dot(electric, num_engines)
        outputs[Dynamic.Mission.NOX_RATE_TOTAL] = np.dot(nox, num_engines)



import numpy as np

from aviary.mission.phase_builder_base import PhaseBuilderBase
from aviary.mission.initial_guess_builders import InitialGuessState, InitialGuessTime, InitialGuessControl
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic
from aviary.mission.gasp_based.ode.ascent_ode import AscentODE


class AscentPhase(PhaseBuilderBase):
    default_name = 'ascent_phase'
    default_ode_class = AscentODE

    _meta_data_ = {}
    _initial_guesses_meta_data_ = {}

    def build_phase(self, aviary_options: AviaryValues = None):
        phase = self.phase = super().build_phase(aviary_options)

        # Retrieve user options values
        user_options = self.user_options

        pitch_constraint_lower = user_options.get_val(
            'pitch_constraint_lower', units='deg')
        pitch_constraint_upper = user_options.get_val(
            'pitch_constraint_upper', units='deg')
        pitch_constraint_ref = user_options.get_val('pitch_constraint_ref', units='deg')
        alpha_constraint_lower = user_options.get_val(
            'alpha_constraint_lower', units='rad')
        alpha_constraint_upper = user_options.get_val(
            'alpha_constraint_upper', units='rad')
        alpha_constraint_ref = user_options.get_val('alpha_constraint_ref', units='rad')

        self.add_flight_path_angle_state(user_options)
        self.add_altitude_state(user_options)
        self.add_velocity_state(user_options)
        self.add_mass_state(user_options)
        self.add_distance_state(user_options, units='ft')

        self.add_altitude_constraint(user_options)

        phase.add_path_constraint(
            "load_factor",
            upper=1.10,
            lower=0.0
        )

        phase.add_path_constraint(
            "fuselage_pitch",
            "theta",
            lower=pitch_constraint_lower,
            upper=pitch_constraint_upper,
            units="deg",
            ref=pitch_constraint_ref,
        )

        phase.add_control(
            "alpha",
            val=0,
            lower=alpha_constraint_lower,
            upper=alpha_constraint_upper,
            units="rad",
            ref=alpha_constraint_ref,
            opt=True,
        )

        phase.add_parameter("t_init_gear", units="s",
                            static_target=True, opt=False, val=38.25)

        phase.add_parameter("t_init_flaps", units="s",
                            static_target=True, opt=False, val=48.21)

        phase.add_timeseries_output(Dynamic.Mission.THRUST_TOTAL, units="lbf")
        phase.add_timeseries_output("normal_force")
        phase.add_timeseries_output(Dynamic.Mission.MACH)
        phase.add_timeseries_output("EAS", units="kn")
        phase.add_timeseries_output(Dynamic.Mission.LIFT)
        phase.add_timeseries_output("CL")
        phase.add_timeseries_output("CD")

        return phase


# Adding metadata for the AscentPhase
AscentPhase._add_meta_data('fix_initial', val=False)
AscentPhase._add_meta_data('angle_lower', val=-15 * np.pi / 180, units='rad')
AscentPhase._add_meta_data('angle_upper', val=25 * np.pi / 180, units='rad')
AscentPhase._add_meta_data('angle_ref', val=np.deg2rad(1), units='rad')
AscentPhase._add_meta_data('angle_ref0', val=0, units='rad')
AscentPhase._add_meta_data('angle_defect_ref', val=0.01, units='rad')
AscentPhase._add_meta_data('alt_lower', val=0, units='ft')
AscentPhase._add_meta_data('alt_upper', val=700, units='ft')
AscentPhase._add_meta_data('alt_ref', val=100, units='ft')
AscentPhase._add_meta_data('alt_ref0', val=0, units='ft')
AscentPhase._add_meta_data('alt_defect_ref', val=100, units='ft')
AscentPhase._add_meta_data('final_altitude', val=500, units='ft')
AscentPhase._add_meta_data('alt_constraint_ref', val=100, units='ft')
AscentPhase._add_meta_data('alt_constraint_ref0', val=0, units='ft')
AscentPhase._add_meta_data('velocity_lower', val=0, units='kn')
AscentPhase._add_meta_data('velocity_upper', val=1000, units='kn')
AscentPhase._add_meta_data('velocity_ref', val=1e2, units='kn')
AscentPhase._add_meta_data('velocity_ref0', val=0, units='kn')
AscentPhase._add_meta_data('velocity_defect_ref', val=None, units='kn')
AscentPhase._add_meta_data('mass_lower', val=0, units='lbm')
AscentPhase._add_meta_data('mass_upper', val=190_000, units='lbm')
AscentPhase._add_meta_data('mass_ref', val=100_000, units='lbm')
AscentPhase._add_meta_data('mass_ref0', val=0, units='lbm')
AscentPhase._add_meta_data('mass_defect_ref', val=1e2, units='lbm')
AscentPhase._add_meta_data('distance_lower', val=0, units='ft')
AscentPhase._add_meta_data('distance_upper', val=10.e3, units='ft')
AscentPhase._add_meta_data('distance_ref', val=3000, units='ft')
AscentPhase._add_meta_data('distance_ref0', val=0, units='ft')
AscentPhase._add_meta_data('distance_defect_ref', val=3000, units='ft')
AscentPhase._add_meta_data('pitch_constraint_lower', val=0, units='deg')
AscentPhase._add_meta_data('pitch_constraint_upper', val=15, units='deg')
AscentPhase._add_meta_data('pitch_constraint_ref', val=1, units='deg')
AscentPhase._add_meta_data('alpha_constraint_lower', val=np.deg2rad(-30), units='rad')
AscentPhase._add_meta_data('alpha_constraint_upper', val=np.deg2rad(30), units='rad')
AscentPhase._add_meta_data('alpha_constraint_ref', val=np.deg2rad(5), units='rad')
AscentPhase._add_meta_data('num_segments', val=None, units='unitless')
AscentPhase._add_meta_data('order', val=None, units='unitless')

# Adding initial guess metadata
AscentPhase._add_initial_guess_meta_data(
    InitialGuessTime(),
    desc='initial guess for time options')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessState('flight_path_angle'),
    desc='initial guess for flight path angle state')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessState('altitude'),
    desc='initial guess for altitude state')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessState('velocity'),
    desc='initial guess for true airspeed state')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessState('mass'),
    desc='initial guess for mass state')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessState('distance'),
    desc='initial guess for distance state')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessControl('alpha'),
    desc='initial guess for angle of attack control')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessControl('tau_gear'),
    desc='when the gear is retracted')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessControl('tau_flaps'),
    desc='when the flaps are retracted')

AscentPhase._add_initial_guess_meta_data(
    InitialGuessControl('throttle'),
    desc='initial guess for throttle')



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials

from aviary.subsystems.aerodynamics.flops_based.induced_drag import InducedDrag
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Dynamic


class InducedDragTest(unittest.TestCase):

    def test_derivs(self):
        P = 2.60239151
        Sref = 1370.0

        CL = np.array([0.3, 0.35, 0.4, 0.45, 0.5, 0.55])
        mach = np.array([0.4, 0.45, 0.5, 0.55, 0.6, 0.85])
        lift = 0.5 * CL * Sref * 1.4 * P * mach ** 2

        nn = len(CL)

        prob = om.Problem(model=om.Group())

        options = {}
        options[Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION] = (False, 'unitless')

        prob.model.add_subsystem('induced_drag', InducedDrag(
            num_nodes=nn, aviary_options=AviaryValues(options)), promotes=['*'])
        prob.setup(force_alloc_complex=True)

        prob.set_val(Dynamic.Mission.MACH, val=mach)
        prob.set_val(Dynamic.Mission.LIFT, val=lift)
        prob.set_val(Dynamic.Mission.STATIC_PRESSURE, val=P)
        prob.set_val(Aircraft.Wing.AREA, val=Sref)
        prob.set_val(Aircraft.Wing.SWEEP, val=-25.03)
        prob.set_val(Aircraft.Wing.TAPER_RATIO, 0.278)
        prob.set_val(Aircraft.Wing.SPAN_EFFICIENCY_FACTOR, 0.7)

        prob.set_val(Aircraft.Wing.ASPECT_RATIO, val=11.05)

        prob.run_model()

        derivs = prob.check_partials(out_stream=None, method="cs")

        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-12, rtol=8e-12)

    def test_derivs_span_eff_redux(self):
        P = 2.60239151
        Sref = 1370.0

        CL = np.array([0.3, 0.35, 0.4, 0.45, 0.5, 0.55])
        mach = np.array([0.4, 0.45, 0.5, 0.55, 0.6, 0.85])
        lift = 0.5 * CL * Sref * 1.4 * P * mach ** 2

        nn = len(CL)

        # High factor

        prob = om.Problem(model=om.Group())

        options = {}
        options[Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION] = (True, 'unitless')

        prob.model.add_subsystem('drag', InducedDrag(
            num_nodes=nn, aviary_options=AviaryValues(options)), promotes=['*'])
        prob.setup(force_alloc_complex=True)

        prob.set_val(Dynamic.Mission.MACH, val=mach)
        prob.set_val(Dynamic.Mission.LIFT, val=lift)
        prob.set_val(Dynamic.Mission.STATIC_PRESSURE, val=P)
        prob.set_val(Aircraft.Wing.AREA, val=Sref)
        prob.set_val(Aircraft.Wing.SWEEP, val=-25.10)
        prob.set_val(Aircraft.Wing.TAPER_RATIO, 0.312)
        prob.set_val(Aircraft.Wing.SPAN_EFFICIENCY_FACTOR, 0.528)

        prob.set_val(Aircraft.Wing.ASPECT_RATIO, val=11.05)

        prob.run_model()

        derivs = prob.check_partials(out_stream=None, method="cs")

        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-12, rtol=8e-12)

        # Low factor.

        prob = om.Problem(model=om.Group())

        options = {}
        options[Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION] = (True, 'unitless')

        prob.model.add_subsystem('drag', InducedDrag(
            num_nodes=nn, aviary_options=AviaryValues(options)), promotes=['*'])
        prob.setup(force_alloc_complex=True)

        prob.set_val(Dynamic.Mission.MACH, val=mach)
        prob.set_val(Dynamic.Mission.LIFT, val=lift)
        prob.set_val(Dynamic.Mission.STATIC_PRESSURE, val=P)
        prob.set_val(Aircraft.Wing.AREA, val=Sref)
        prob.set_val(Aircraft.Wing.SWEEP, val=-25.10)
        prob.set_val(Aircraft.Wing.TAPER_RATIO, 0.312)
        prob.set_val(Aircraft.Wing.SPAN_EFFICIENCY_FACTOR, 0.528)

        prob.set_val(Aircraft.Wing.ASPECT_RATIO, val=11.05)

        prob.run_model()

        derivs = prob.check_partials(out_stream=None, method="cs")

        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-12, rtol=8e-12)


if __name__ == "__main__":
    unittest.main()



import unittest
import os

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.aerodynamics.gasp_based.flaps_model.L_and_D_increments import \
    LiftAndDragIncrements
from aviary.variable_info.variables import Aircraft

"""
All data is from validation files using standalone flaps model
"""


class LiftAndDragIncrementsTestCase(unittest.TestCase):
    def setUp(self):

        self.prob = om.Problem(model=om.Group())

        self.prob.model.add_subsystem('LaDIs', LiftAndDragIncrements(), promotes=['*'])

        self.prob.setup()

        # initial conditions
        self.prob.set_val(Aircraft.Wing.FLAP_DRAG_INCREMENT_OPTIMUM, 0.1)
        self.prob.set_val(Aircraft.Wing.FLAP_LIFT_INCREMENT_OPTIMUM, 1.5)
        self.prob.set_val("VDEL1", 1.0)
        self.prob.set_val("VDEL2", 1)
        self.prob.set_val("VDEL3", 0.765)
        self.prob.set_val("VDEL4", 0.93578)
        self.prob.set_val("VDEL5", 0.90761)
        self.prob.set_val("VLAM3", 0.97217)
        self.prob.set_val("VLAM4", 1.25725)
        self.prob.set_val("VLAM5", 1.0)
        self.prob.set_val("VLAM6", 1.0)
        self.prob.set_val("VLAM7", 0.735)
        self.prob.set_val("VLAM8", 0.74444)
        self.prob.set_val("VLAM13", 1.03512)
        self.prob.set_val("VLAM14", 0.99124)

    def test_case(self):

        self.prob.run_model()
        tol = 5e-4
        print()

        reg_data = 0.0650
        ans = self.prob["delta_CD"]
        assert_near_equal(ans, reg_data, tol)

        reg_data = 1.0293
        ans = self.prob["delta_CL"]
        assert_near_equal(ans, reg_data, tol)

        data = self.prob.check_partials(out_stream=None, method="fd")
        assert_check_partials(data, atol=1e-4, rtol=1e-4)


if __name__ == "__main__":
    unittest.main()



from enum import Enum, auto, unique


class AlphaModes(Enum):
    '''
    AlphaModes is used to specify how angle of attack is defined during
        climb and descent.
    DEFAUT:
        Alpha is an input
    ROTATION
        Alpha is calculated as the initial angle plus the rotation rate
        times the duration of the rotation.
    LOAD_FACTOR
        Alpha is limited to ensure the load factor never exceeds a
        specified maximum.
    FUSELAGE_PITCH
        Alpha is calculated to set a particular floor angle given the 
        current flight path angle.
    DECELERATION
        Alpha is calculated to target a specified TAS rate, the default
        is a TAS rate of 0 (Constant TAS).
    REQUIRED_LIFT
        Alpha is calculated such that the aircraft produces a particular
        lifting force.
    ALTITUDE_RATE
        Alpha is calculated to target a specified altitude rate, the default
        is 0 (Constant Altitude).
    '''
    DEFAULT = auto()
    ROTATION = auto()
    LOAD_FACTOR = auto()
    FUSELAGE_PITCH = auto()
    DECELERATION = auto()
    REQUIRED_LIFT = auto()
    ALTITUDE_RATE = auto()
    CONSTANT_ALTITUDE = auto()


class AnalysisScheme(Enum):
    """
    AnalysisScheme is used to select from Collocation and shooting.

    COLLOCATION uses the collocation method to optimize all points simultaneously
    and can be run in parallel. However, it requires reasonable initial guesses
    for the trajectory and is fairly sensitive to those initial guesses.

    SHOOTING is a forward in time integration method that simulates the trajectory.
    This does not require initial guesses and will always produce physically valid
    trajectories, even during optimizer failures. The shooting method cannot be run
    in parallel.
    """
    COLLOCATION = auto()
    SHOOTING = auto()


class EquationsOfMotion(Enum):
    """
    Available equations of motion for use during mission analysis
    """
    HEIGHT_ENERGY = 'height_energy'
    TWO_DEGREES_OF_FREEDOM = '2DOF'
    SOLVED = 'solved'


@unique
class GASPEngineType(Enum):
    """
    Defines the type of engine to use in GASP-based mass calculations.
    Note that only the value for the first engine model will be used.
    Currenly only the TURBOJET option is implemented, but other types of engines will be added in the future.
    """
    # Reciprocating engine with carburator
    RECIP_CARB = 1

    # Reciprocating engine with fuel injection
    RECIP_FUEL_INJECT = 2

    # Reciprocating engine with fuel injection and geared
    RECIP_FUEL_INJECT_GEARED = 3

    # Rotary-combustion engine
    ROTARY = 4

    # Turboshaft engine
    TURBOSHAFT = 5

    # Turboprop engine
    TURBOPROP = 6

    # Turbojet or turbofan engine
    TURBOJET = 7

    # Reciprocating engine with carburator; use HOPWSZ (horizontally-opposed piston
    # weight and size) methodology for geometry and mass
    RECIP_CARB_HOPWSZ = 11

    # Reciprocating engine with fuel injection; use HOPWSZ (horizontally-opposed piston
    # weight and size) methodology for geometry and mass
    RECIP_FUEL_INJECT_HOPWSZ = 12

    # Reciprocating engine with fuel injection and geared; use HOPWSZ (horizontally-
    # opposed piston weight and size) methodology for geometry and mass
    RECIP_FUEL_INJECT_GEARED_HOPWSZ = 13

    # Rotary-combustion engine; use RCWSZ (rotary combustion weight and size) methodology
    # for geometry and mass
    ROTARY_RCWSZ = 14


@unique
class FlapType(Enum):
    """
    Defines the type of flap used on the wing. Used in GASP-based aerodynamics and mass calculations.
    """
    PLAIN = 1
    SPLIT = 2
    SINGLE_SLOTTED = 3
    DOUBLE_SLOTTED = 4
    TRIPLE_SLOTTED = 5
    FOWLER = 6
    DOUBLE_SLOTTED_FOWLER = 7


class LegacyCode(Enum):
    """
    Flag for legacy codebases
    """
    FLOPS = 'FLOPS'
    GASP = 'GASP'

    def __str__(self):
        return self.value


class ProblemType(Enum):
    """
    ProblemType is used to switch between different combinations of
    design variables and constraints.

    SIZING: Varies the design gross weight and actual gross weight to
    close to design range. This causes the empty weight and the fuel
    weight to change.

    ALTERNATE: Requires a pre-sized aircraft. It holds the design gross
    weight and empty weight constant. It then varies the fuel weight
    and actual gross weight until the range closes to the off-design
    range.

    FALLOUT: Requires a pre-sized aircraft. It holds the design gross
    weight and empty weight constant. Using the specified actual
    gross weight, it will then find the maximum distance the off-design
    aircraft can fly.
    """
    SIZING = 'sizing'
    ALTERNATE = 'alternate'
    FALLOUT = 'fallout'


class SpeedType(Enum):
    '''
    SpeedType is used to specify the type of speed being used.
    EAS is equivalent airspeed.
    TAS is true airspeed.
    MACH is mach
    '''
    EAS = auto()
    TAS = auto()
    MACH = auto()


class Verbosity(Enum):
    """
    Sets how much information Aviary outputs when run

    Verbosity levels are based on ubuntu's standard:
    https://discourse.ubuntu.com/t/cli-verbosity-levels/26973
    """
    QUIET = 0
    BRIEF = 1
    VERBOSE = 2
    DEBUG = 3



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.mass_to_weight import MassToWeight


class MassToWeightTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    def test_case(self):

        prob = om.Problem()

        prob.model.add_subsystem(
            "calc_weight",
            MassToWeight(),
            promotes=['mass', 'weight']
        )

        prob.setup(force_alloc_complex=True)

        prob.set_val('mass', 120_000, 'lbm')

        prob.run_model()
        prob.model.list_inputs()
        prob.model.list_outputs()
        assert_near_equal(prob.get_val('weight', 'lbf'),
                          120_000 * GRAV_ENGLISH_LBM, 1.0e-12)
        partial_data = prob.check_partials(compact_print=True, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-12)

        nn = 3
        mass = np.full(nn, 120_000)

        prob = om.Problem()

        prob.model.add_subsystem(
            "calc_weight",
            MassToWeight(num_nodes=nn),
            promotes=['mass', 'weight']
        )

        prob.setup(force_alloc_complex=True)

        prob.set_val('mass', mass, 'lbm')

        prob.run_model()
        prob.model.list_inputs()
        prob.model.list_outputs()
        assert_near_equal(prob.get_val('weight', 'lbf'),
                          mass * GRAV_ENGLISH_LBM, 1.0e-12)
        partial_data = prob.check_partials(compact_print=True, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-12)


if __name__ == '__main__':
    unittest.main()



import unittest

import aviary.api as av
from aviary.utils.merge_variable_metadata import merge_meta_data

dict1 = av.CoreMetaData.copy()
# this is the baseline Aviary-core metadata
dict2 = av.CoreMetaData.copy()
# This is the baseline Aviary-core metadata with a few variables added onto it
dict3 = av.CoreMetaData.copy()
# This is the baseline Aviary-core metadata with a few variables added onto it, including some duplicates from dict2 that have the same metadata
dict4 = av.CoreMetaData.copy()
# This is the baseline Aviary-core metadata with a few variables added onto it, include a duplicate from dict2 that has different metadata
merged_dicts_23 = av.CoreMetaData.copy()
# This is the expected result of merging dict2 with dict3

############################################ Make dict2 ############################################
av.add_meta_data(
    'aircraft:CG',
    units='ft',
    desc='Center of gravity',
    default_value=0,
    meta_data=dict2,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:wing:character',
    units=None,
    desc='Character of wing',
    default_value=0,
    meta_data=dict2,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:landing_gear:shape',
    units='ft',
    desc='Shape of landing gear',
    default_value=0,
    meta_data=dict2,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:engine:color',
    units=None,
    desc='Color of engine',
    default_value='red',
    meta_data=dict2,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:fuselage:eccentricity',
    units=None,
    desc='Eccentricity of cross section of fuselage',
    default_value=.5,
    meta_data=dict2,
    historical_name=None,
)


############################################ Make dict3 ############################################
av.add_meta_data(
    'aircraft:CG',  # This is a duplicate from dict2 with the same metadata
    units='ft',
    desc='Center of gravity',
    default_value=0,
    meta_data=dict3,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:wing:temperature',
    units='degK',
    desc='Temperature of the wing on the ground',
    default_value=273,
    meta_data=dict3,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:strut:superiority',
    units=None,
    desc='Amount by which the strut is better than having no strut',
    default_value=2,
    meta_data=dict3,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:engine:color',  # This is a duplicate from dict2 with the same metadata
    units=None,
    desc='Color of engine',
    default_value='red',
    meta_data=dict3,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:overall_color',
    units=None,
    desc='The predominant color of the vehicle',
    default_value='purple',
    meta_data=dict3,
    historical_name=None,
)


############################################ Make dict4 ############################################
av.add_meta_data(
    'aircraft:CG',  # This is a duplicate from dict2 with the same metadata
    units='ft',
    desc='Center of gravity',
    default_value=0,
    meta_data=dict4,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:wing:temperature',
    units='degK',
    desc='Temperature of the wing on the ground',
    default_value=273,
    meta_data=dict4,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:strut:superiority',
    units=None,
    desc='Amount by which the strut is better than having no strut',
    default_value=2,
    meta_data=dict4,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:engine:color',  # This is a duplicate from dict2 with different metadata
    units=None,
    desc='Color of enigne',  # note there is a subtle typo in the description here
    default_value='red',
    meta_data=dict4,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:overall_color',
    units=None,
    desc='The predominant color of the vehicle',
    default_value='purple',
    meta_data=dict4,
    historical_name=None,
)


############################################ Make merge of dicts 2&3 ############################################
av.add_meta_data(
    'aircraft:CG',
    units='ft',
    desc='Center of gravity',
    default_value=0,
    meta_data=merged_dicts_23,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:wing:character',
    units=None,
    desc='Character of wing',
    default_value=0,
    meta_data=merged_dicts_23,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:landing_gear:shape',
    units='ft',
    desc='Shape of landing gear',
    default_value=0,
    meta_data=merged_dicts_23,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:engine:color',
    units=None,
    desc='Color of engine',
    default_value='red',
    meta_data=merged_dicts_23,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:fuselage:eccentricity',
    units=None,
    desc='Eccentricity of cross section of fuselage',
    default_value=.5,
    meta_data=merged_dicts_23,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:wing:temperature',
    units='degK',
    desc='Temperature of the wing on the ground',
    default_value=273,
    meta_data=merged_dicts_23,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:strut:superiority',
    units=None,
    desc='Amount by which the strut is better than having no strut',
    default_value=2,
    meta_data=merged_dicts_23,
    historical_name=None,
)

av.add_meta_data(
    'aircraft:overall_color',
    units=None,
    desc='The predominant color of the vehicle',
    default_value='purple',
    meta_data=merged_dicts_23,
    historical_name=None,
)
############################################ Make merge of dicts 1&2&3 ############################################
# This is the expected result of merging dict1 with dict2 and dict3
merged_dicts_123 = merged_dicts_23.copy()
############################################ Make merge of dicts 1&2 ############################################
merged_dicts_12 = dict2.copy()  # This is the expected result of merging dict1 with dict2
############################################ Make merge of dicts 1&3 ############################################
merged_dicts_13 = dict3.copy()  # This is the expected result of merging dict1 with dict3
############################################ Make expected error message for attempted merge of dicts 2&4 ############################################
# This is the expected error message resulting from mergin dict2 with dict4
merged_dicts_24_msg = 'You have attempted to merge metadata dictionaries that contain the same variable with different metadata. The offending variable present in multiple dictionaries is "aircraft:engine:color".'

############################################ Make dict5 ############################################
dict5 = av.CoreMetaData.copy()

av.add_meta_data(
    'aircraft:fuselage:eccentricity',
    units=None,
    desc='Eccentricity of cross section of fuselage',
    default_value=0.500000001,  # a small but acceptable difference
    meta_data=dict5,
    historical_name=None,
)

############################################ Make dict6 ############################################
dict6 = av.CoreMetaData.copy()

av.add_meta_data(
    'aircraft:fuselage:eccentricity',
    units=None,
    desc='Eccentricity of cross section of fuselage',
    default_value=0.50001,  # a larger difference
    meta_data=dict6,
    historical_name=None,
)

# expected error message for merging dict2 and dict6
merged_dicts_26_msg = 'You have attempted to merge metadata dictionaries that contain the same variable with different metadata. The offending variable present in multiple dictionaries is "aircraft:fuselage:eccentricity".'


class MergeMetaDataTest(unittest.TestCase):
    def test_match_merge(self):
        merge23 = merge_meta_data([dict2, dict3])
        merge123 = merge_meta_data([dict1, dict2, dict3])
        merge12 = merge_meta_data([dict1, dict2])
        merge13 = merge_meta_data([dict1, dict3])
        # test the small, acceptable difference
        merge25 = merge_meta_data([dict2, dict5])

        self.assertEqual(merge23, merged_dicts_23)
        self.assertEqual(merge123, merged_dicts_123)
        self.assertEqual(merge12, merged_dicts_12)
        self.assertEqual(merge13, merged_dicts_13)
        self.assertEqual(merge25, dict2)  # dict5 is essentially the same as dict2

    # ensure that the correct error message is generated when two dictionaries are merged that have the same variable with different metadata
    def test_mismatch_error(self):
        with self.assertRaises(ValueError) as cm:
            merge_meta_data([dict2, dict4])
        self.assertEqual(str(cm.exception), merged_dicts_24_msg)

        with self.assertRaises(ValueError) as cm:
            merge_meta_data([dict2, dict6])  # test the larger difference
        self.assertEqual(str(cm.exception), merged_dicts_26_msg)


if __name__ == '__main__':
    unittest.main()



import unittest
import copy

from aviary.interface.utils.check_phase_info import check_phase_info
from aviary.interface.default_phase_info.height_energy import phase_info as phase_info_height_energy
from aviary.interface.default_phase_info.two_dof import phase_info as phase_info_two_dof
from aviary.variable_info.enums import EquationsOfMotion

HEIGHT_ENERGY = EquationsOfMotion.HEIGHT_ENERGY
TWO_DEGREES_OF_FREEDOM = EquationsOfMotion.TWO_DEGREES_OF_FREEDOM


class TestCheckInputs(unittest.TestCase):
    def test_correct_input_height_energy(self):
        # This should pass without any issue as it's the same valid dict as before
        self.assertTrue(check_phase_info(
            phase_info_height_energy, mission_method=HEIGHT_ENERGY))

    def test_incorrect_tuple(self):
        # Let's replace a tuple with a float in the 'climb' phase
        incorrect_tuple_phase_info = copy.deepcopy(phase_info_height_energy)
        incorrect_tuple_phase_info['climb']['user_options']['initial_altitude'] = 10.668e3
        with self.assertRaises(ValueError):
            check_phase_info(incorrect_tuple_phase_info, mission_method=HEIGHT_ENERGY)

    def test_incorrect_unit(self):
        # Let's replace a valid unit with an invalid one in the 'climb' phase
        incorrect_unit_phase_info = copy.deepcopy(phase_info_height_energy)
        incorrect_unit_phase_info['climb']['user_options']['initial_altitude'] = (
            10.668e3, 'invalid_unit')
        with self.assertRaisesRegex(ValueError,
                                    "Key initial_altitude in phase climb has an invalid unit invalid_unit."):
            check_phase_info(incorrect_unit_phase_info, mission_method=HEIGHT_ENERGY)

    def test_correct_input_two_dof(self):
        # This should pass without any issue as it's the same valid dict as before
        self.assertTrue(check_phase_info(phase_info_two_dof,
                        mission_method=TWO_DEGREES_OF_FREEDOM))

    def test_incorrect_mission_method(self):
        # Let's pass an incorrect mission_method name
        incorrect_mission_method = 'INVALID_METHOD'
        with self.assertRaises(ValueError):
            check_phase_info(phase_info_height_energy,
                             mission_method=incorrect_mission_method)


if __name__ == '__main__':
    unittest.main()



"""
Top level run script for an Aviary problem.

This theoretically could be the Level 2 (intermediate) user's
entry point to Aviary.
"""

from aviary.api import AviaryProblem
from aviary.api import AnalysisScheme, SpeedType, AlphaModes
from aviary.api import FlexibleTraj
from aviary.api import create_2dof_based_ascent_phases
from aviary.api import SGMCruise, SGMDescent
from aviary.api import Dynamic


def run_aviary(aircraft_filename, phase_info, optimizer=None, analysis_scheme=AnalysisScheme.COLLOCATION,
               objective_type=None, record_filename='dymos_solution.db', restart_filename=None, max_iter=50, run_driver=True, make_plots=True):
    """
    This function runs the aviary optimization problem for the specified aircraft configuration and mission.

    It first creates an instance of the AviaryProblem class using the given phase_info, mission_method, and mass_method.
    It then loads aircraft and options data from the user-provided aircraft_filename and checks for any clashing inputs.
    Pre-mission systems are added, phases are added and linked, then post-mission systems are added.
    A driver is added using the specified optimizer (The default optimizer depends on the analysis scheme).
    Design variables and the optimization objective are added, and the problem is set up.
    Initial guesses are set and the optimization problem is run.
    The function returns the AviaryProblem instance.

    A user can modify these methods or add their own to modify the behavior
    of the Aviary problem.
    """
    # Build problem
    prob = AviaryProblem(analysis_scheme)

    # Load aircraft and options data from user
    # Allow for user overrides here
    prob.load_inputs(aircraft_filename, phase_info)


# Preprocess inputs
    prob.check_and_preprocess_inputs()

    prob.add_pre_mission_systems()

# ######################################## #
# replace the default trajectory with a custom trajectory
# This trajectory uses the full GASP based ascent profile,
# a Breguet cruise, and a simplified descent
    ascent_phases = create_2dof_based_ascent_phases(
        prob.ode_args,
        cruise_alt=prob.cruise_alt,
        cruise_mach=prob.cruise_mach)

    cruise_kwargs = dict(
        input_speed_type=SpeedType.MACH,
        input_speed_units="unitless",
        ode_args=prob.ode_args,
        alpha_mode=AlphaModes.REQUIRED_LIFT,
        simupy_args=dict(
            DEBUG=True,
        ),
    )
    cruise_vals = {
        'mach': {'val': prob.cruise_mach, 'units': cruise_kwargs['input_speed_units']},
        'distance_trigger': {'val': 300, 'units': 'NM'},
    }

    descent1_kwargs = dict(
        input_speed_type=SpeedType.MACH,
        input_speed_units="unitless",
        speed_trigger_units='kn',
        ode_args=prob.ode_args,
        simupy_args=dict(
            DEBUG=False,
        ),
    )
    descent1_vals = {
        'alt_trigger': {'val': 10000, 'units': 'ft'},
        'mach': {'val': prob.cruise_mach, 'units': None},
        'speed_trigger': {'val': 350, 'units': 'kn'}}

    phases = {
        **ascent_phases,
        'cruise': {
            'ode': SGMCruise(**cruise_kwargs),
            'vals_to_set': cruise_vals,
        },
        'descent1': {
            'ode': SGMDescent(**descent1_kwargs),
            'vals_to_set': descent1_vals
        },
    }

    traj = FlexibleTraj(
        Phases=phases,
        traj_final_state_output=[
            Dynamic.Mission.MASS,
            Dynamic.Mission.DISTANCE,
        ],
        traj_initial_state_input=[
            Dynamic.Mission.MASS,
            Dynamic.Mission.DISTANCE,
            Dynamic.Mission.ALTITUDE,
        ],
        traj_event_trigger_input=[
            (phases['groundroll']['ode'], Dynamic.Mission.VELOCITY, 0,),
            (phases['climb3']['ode'], Dynamic.Mission.ALTITUDE, 0,),
            (phases['cruise']['ode'], Dynamic.Mission.MASS, 0,),
        ],
    )
    traj = prob.model.add_subsystem('traj', traj)

    prob.traj = traj
# ######################################## #

    prob.add_post_mission_systems()

    # Link phases and variables
    prob.link_phases()

    prob.add_driver(optimizer, max_iter=max_iter)

    prob.add_design_variables()

    # Load optimization problem formulation
    # Detail which variables the optimizer can control
    prob.add_objective(objective_type=objective_type)

    prob.setup()

    prob.set_initial_guesses()

    prob.failed = prob.run_aviary_problem(
        record_filename, restart_filename=restart_filename, run_driver=run_driver, make_plots=make_plots)

    return prob


if __name__ == "__main__":
    from aviary.interface.default_phase_info.two_dof import phase_info
    input_deck = 'models/large_single_aisle_1/large_single_aisle_1_GwGm.csv'
    run_aviary(input_deck, phase_info,
               analysis_scheme=AnalysisScheme.SHOOTING, run_driver=False)



#!/usr/bin/python

import argparse
import getpass
import itertools
from datetime import datetime
from enum import Enum
from pathlib import Path

import numpy as np
import openmdao.api as om
from dymos.models.atmosphere import USatm1976Comp
from openmdao.components.interp_util.interp import InterpND

from aviary.subsystems.propulsion.engine_deck import normalize
from aviary.subsystems.propulsion.utils import EngineModelVariables, default_units
from aviary.variable_info.variables import Dynamic
from aviary.utils.csv_data_file import write_data_file
from aviary.utils.functions import get_path
from aviary.utils.named_values import NamedValues


class EngineDeckType(Enum):
    FLOPS = 'FLOPS'
    GASP = 'GASP'


MACH = EngineModelVariables.MACH
ALTITUDE = EngineModelVariables.ALTITUDE
THROTTLE = EngineModelVariables.THROTTLE
HYBRID_THROTTLE = EngineModelVariables.HYBRID_THROTTLE
THRUST = EngineModelVariables.THRUST
GROSS_THRUST = EngineModelVariables.GROSS_THRUST
RAM_DRAG = EngineModelVariables.RAM_DRAG
FUEL_FLOW = EngineModelVariables.FUEL_FLOW
ELECTRIC_POWER = EngineModelVariables.ELECTRIC_POWER
NOX_RATE = EngineModelVariables.NOX_RATE
TEMPERATURE = EngineModelVariables.TEMPERATURE_ENGINE_T4
# EXIT_AREA = EngineModelVariables.EXIT_AREA

flops_keys = [
    MACH,
    ALTITUDE,
    THROTTLE,
    GROSS_THRUST,
    RAM_DRAG,
    FUEL_FLOW,
    NOX_RATE]  # , EXIT_AREA]

# later code assumes T4 is last item in keys
gasp_keys = [MACH, ALTITUDE, THROTTLE, THRUST, FUEL_FLOW, TEMPERATURE]

header_names = {
    MACH: 'Mach_Number',
    ALTITUDE: 'Altitude',
    THROTTLE: 'Throttle',
    THRUST: 'Thrust',
    GROSS_THRUST: 'Gross_Thrust',
    RAM_DRAG: 'Ram_Drag',
    FUEL_FLOW: 'Fuel_Flow',
    NOX_RATE: 'NOx_Rate',
    TEMPERATURE: 'T4',
    # EXIT_AREA: 'Exit Area',
}


def EngineDeckConverter(input_file=None, output_file=None, data_format=None):
    '''
    Converts FLOPS- or GASP-formatted engine decks into Aviary csv format.
    FLOPS decks are changed from column-delimited to csv format with added headers.
    GASP decks are reorganized into csv. T4 is recovered using assumptions used in GASPy.
    Data points whose T4 exceeds T4max are removed.

    Parameters
    ----------
    input_file : (str, Path)
        path to engine deck file to be converted
    output_file : (str, Path)
        path to file where new converted data will be written
    data_format : (EngineDeckType)
        data format used by input_file (FLOPS or GASP)
    readable : (bool)
        output_file will be organized with consistent column widths for easier reading
    '''
    # TODO rounding for calculated values?

    timestamp = datetime.now().strftime('%m/%d/%y at %H:%M')
    user = getpass.getuser()
    comments = []
    header = {}
    data = {}

    data_file = get_path(input_file)

    comments.append(f'# created {timestamp} by {user}')
    comments.append(
        f'# {data_format.value}-derived engine deck converted from {data_file.name}')
    if data_format == EngineDeckType.FLOPS:
        header = {key: default_units[key] for key in flops_keys}
        data = {key: np.array([]) for key in flops_keys}

        with open(data_file, newline='', encoding='utf-8-sig') as file:
            reader = _read_flops_engine(file)
            data_starts = False

            for line in reader:
                if not data_starts:
                    # pull out comments before data starts
                    if line[0][0] == '#':
                        comments.append(line)
                        continue
                    data_starts = True

                data[MACH] = np.append(data[MACH], line[0])
                data[ALTITUDE] = np.append(data[ALTITUDE], line[1])
                data[THROTTLE] = np.append(data[THROTTLE], line[2])
                data[GROSS_THRUST] = np.append(data[GROSS_THRUST], line[3])
                data[RAM_DRAG] = np.append(data[RAM_DRAG], line[4])
                data[FUEL_FLOW] = np.append(data[FUEL_FLOW], line[5])
                data[NOX_RATE] = np.append(data[NOX_RATE], line[6])
                # data[EXIT_AREA].append(line[7])

    elif data_format == EngineDeckType.GASP:
        data = {key: [] for key in gasp_keys}

        scalars, tables = _read_gasp_engine(data_file)
        # save scalars as comments
        comments.extend(['# ' + key + ': ' + str(scalars[key])
                        for key in scalars.keys()])

        # recommended to always generate structured grid
        structure_data = True
        if structure_data:
            structured_data = _make_structured_grid(tables, method='lagrange3')

            data[MACH] = structured_data['thrust']['machs']
            data[ALTITUDE] = structured_data['thrust']['alts']
            data[THRUST] = structured_data['thrust']['vals']
            data[FUEL_FLOW] = structured_data['fuelflow']['vals']
            T4T2 = structured_data['thrust']['t4t2s']

        else:
            data[MACH] = tables['thrust'][:, 2]
            data[ALTITUDE] = tables['thrust'][:, 0]
            data[THRUST] = tables['thrust'][:, 3]
            data[FUEL_FLOW] = tables['fuelflow'][:, 3]
            T4T2 = tables['thrust'][:, 1]

        generate_flight_idle = True
        if generate_flight_idle:
            data, T4T2 = _generate_flight_idle(data, T4T2,
                                               ref_sls_airflow=scalars['sls_airflow'],
                                               ref_sfn_idle=scalars['sfn_idle'])

        t4max = scalars['t4max']

        # if t4max 100 or less, it is actually throttle. Remove temperature as variable
        if t4max <= 100:
            compute_T4 = False
            data.pop(TEMPERATURE)
            # temperature is assumed last in keys
            gasp_keys.pop(-1)
        else:
            compute_T4 = True

        # define header now that we know what is in the engine deck
        header = {key: default_units[key] for key in gasp_keys}

        if compute_T4:
            # compute T4 using atmospheric model
            prob = om.Problem()

            prob.model.add_subsystem('T4T2', om.IndepVarComp('T4:T2',
                                                             T4T2,
                                                             units='unitless'),
                                     promotes=['*'])

            prob.model.add_subsystem(
                Dynamic.Mission.MACH,
                om.IndepVarComp(
                    Dynamic.Mission.MACH,
                    data[MACH],
                    units='unitless'),
                promotes=['*'])

            prob.model.add_subsystem(
                Dynamic.Mission.ALTITUDE,
                om.IndepVarComp(
                    Dynamic.Mission.ALTITUDE,
                    data[ALTITUDE],
                    units='ft'),
                promotes=['*'])

            prob.model.add_subsystem(
                name='atmosphere',
                subsys=USatm1976Comp(num_nodes=len(data[MACH])),
                promotes_inputs=[('h', Dynamic.Mission.ALTITUDE)],
                promotes_outputs=[('temp', Dynamic.Mission.TEMPERATURE)])

            prob.model.add_subsystem(
                name='conversion',
                subsys=AtmosCalc(num_nodes=len(data[MACH])),
                promotes_inputs=[Dynamic.Mission.MACH,
                                 Dynamic.Mission.TEMPERATURE],
                promotes_outputs=['t2']
            )

            prob.setup()
            prob.run_model()

            T2 = prob.get_val('t2')
            T4 = T2 * T4T2
            data[TEMPERATURE] = T4
            # Throttle is T4 normalized from 0 to 1 (T4max)
            # By always keeping minimum T4 zero for normalization, throttle stays
            #   consistent with fraction of T4max
            # TODO flight condition dependent throttle range?
            # NOTE this often leaves max throttles less than 1 in the deck - this caues
            #     problems when finding reference SLS thrust, as there is often no max
            #     power data at that point in the engine deck. It is reccomended GASP
            #     engine decks override Aircraft.Engine.REFERENCE_THRUST in EngineDecks
            data[THROTTLE] = normalize(data[TEMPERATURE], minimum=0.0, maximum=t4max)

            # remove all points above T4max
            # TODO save these points as commented out?
            valid_idx = np.where(data[THROTTLE] <= 1.0)
            data[MACH] = data[MACH][valid_idx]
            data[ALTITUDE] = data[ALTITUDE][valid_idx]
            data[THROTTLE] = data[THROTTLE][valid_idx]
            data[THRUST] = data[THRUST][valid_idx]
            data[FUEL_FLOW] = data[FUEL_FLOW][valid_idx]
            data[TEMPERATURE] = data[TEMPERATURE][valid_idx]

        else:
            data[THROTTLE] = T4T2

        # data needs to be string so column length can be easily found later
        for var in data:
            data[var] = np.array([str(item) for item in data[var]])

    else:
        quit("Invalid engine deck format provided")

    # sort data
    # create parallel dict to data that stores floats
    formatted_data = {}
    for key in data:
        formatted_data[key] = data[key].astype(float)

    # convert engine_data from dict to list so it can be sorted
    sorted_values = np.array([formatted_data[key] for key in formatted_data]).transpose()

    # Sort by mach, then altitude, then throttle, then hybrid throttle
    sorted_values = sorted_values[np.lexsort(
        [formatted_data[THROTTLE],
         formatted_data[ALTITUDE],
         formatted_data[MACH]])]
    for idx, key in enumerate(formatted_data):
        formatted_data[key] = sorted_values[:, idx]

    # rework formatted_data into 2d array
    formatted_data = np.array([formatted_data[key]
                              for key in header.keys()]).transpose().astype(float)

    # store formatted data into NamedValues object
    write_data = NamedValues()
    for idx, key in enumerate(data):
        write_data.set_val(header_names[key], data[key], default_units[key])

    write_data_file(output_file, write_data, comments, include_timestamp=False)


def _read_flops_engine(input_file):
    '''
    Read engine data file using FLOPS standard, which is column delimited data
    always assumed to be in the order defined in the FLOPS manual
    '''
    for line in input_file:
        sz = len(line)

        if sz < 80:
            line = line + (80 - sz) * ' '

        if line[0] == '#':
            data = line.strip()
        else:
            data = [
                _flops_field_convert(line[0:5]),
                _flops_field_convert(line[5:15]),
                _flops_field_convert(line[15:20]),
                _flops_field_convert(line[20:30]),
                _flops_field_convert(line[30:40]),
                _flops_field_convert(line[40:50]),
                # intenional gap from 50:60 - column is left blank in FLOPS standard
                _flops_field_convert(line[60:70]),
                _flops_field_convert(line[70:80]),
            ]

        yield data


def _flops_field_convert(arg: str):
    rvalue = arg.strip()

    if not rvalue:
        rvalue = _flops_empty_field

    return rvalue


# in FLOPS, empty fields are converted to zero
_flops_empty_field = '0'


def _read_gasp_engine(fp):
    """Read a GASP engine deck file and parse its scalars and tabular data.
    Scalars (T4 max, SLS airflow, etc.) are read from the first line of the engine deck
    (IREAD=1 is assumed) and returned in a dictionary.
    Data tables are also returned as a dictionary, with separate tables for thrust,
    fuelflow, and airflow, since they may have different grids in general.
    Each table consists of both the independent variables and the dependent variable for
    the corresponding field. The table is a "tidy format" 2D array where the first three
    columns are the independent varaiables (altitude, T4/T2, and Mach number) and the
    final column is the dependent variable (one of thrust, fuelflow, or airflow).
    """
    with open(fp, "r") as f:
        scalars = _read_header(f)
        tables = {k: _read_table(f) for k in ["thrust", "fuelflow", "airflow"]}

    return scalars, tables


def _read_header(f):
    """Read GASP engine deck header, returning the engine scalars in a dict"""
    # file header: FORMAT(2I5,10X,5F10.4)
    iread, iprint, wamap, t4max, t4mcl, t4mc, sfnidl = _parse(
        f, [*_rep(2, (int, 5)), (None, 10), *_rep(5, (float, 10))]
    )

    if iread != 1:
        raise RuntimeError(f"IREAD=1 expected, got {iread}")

    return {
        "t4max": t4max,
        "t4cruise": t4mc,
        "t4climb": t4mcl,
        "sls_airflow": wamap,
        "sfn_idle": sfnidl,
    }


def _read_table(f):
    """Read an entire table from a GASP engine deck file.
    The table data is returned as a "tidy format" array with three columns for the
    independent variables (altitude, T4/T2, and Mach number) and the final column for
    the table field (one of thrust, fuelflow, or airflow).
    """
    tab_data = None

    # table title
    title = f.readline().strip()
    # number of maps in the table
    (nmaps,) = _parse(f, [(int, 5)])
    # blank line
    f.readline()

    for i in range(nmaps):
        map_data = _read_map(f)

        # blank line following all but the last map in the table
        if i < nmaps - 1:
            f.readline()

        if tab_data is None:
            tab_data = map_data
        else:
            tab_data = np.r_[tab_data, map_data]

    return tab_data


def _rep(n, t):
    """Shorthand for ``itertools.repeat`` with the multiplier first."""
    return itertools.repeat(t, n)


def _parse(f, fmt):
    """Read a line from file ``f`` and parse it according to the given ``fmt``"""
    return _strparse(f.readline(), fmt)


def _strparse(s, fmt):
    """Parse a string into fixed-width numeric fields.
    ``fmt`` should be a list of tuples specifying (type, length) for each field in
    string ``s``. Use None for the type to skip (i.e. not yield) that field.
    """
    p = 0
    for typ, length in fmt:
        sub = s[p: p + length]
        if typ is not None:
            yield typ(sub)
        p += length


def _read_map(f):
    """Read a single map of a table from the engine deck file.
    The map data is returned in the same format as in ``read_table``, except there is a
    single altitude value per map.
    """
    # map dimensions: FORMAT(/2I5,F10.1,10X))
    npts, nline, amap = _parse(f, [*_rep(2, (int, 5)), (float, 10)])

    map_data = np.empty((npts * nline, 4))
    map_data[:, 0] = amap

    # number of points on a single line - wrapped if more than 6
    nptloc = min(6, npts)

    # point vals: FORMAT(10X,6F10.4,10X)
    x = list(_parse(f, [(None, 10), *_rep(nptloc, (float, 10))]))
    if npts > nptloc:
        # remaining vals on wrapped line
        x.extend(list(_parse(f, [(None, 10), *_rep(npts - nptloc, (float, 10))])))

    map_data[:, 2] = np.tile(x, nline)

    for j in range(nline):
        # line (y) val then z vals: FORMAT(F10.4,6F10.1,10X,/(6F10.1,10X))
        vals = list(_parse(f, [(float, 10), *_rep(nptloc, (float, 10))]))
        y = vals[0]
        z = vals[1:]
        if npts > nptloc:
            # ad remaining vals on warapped line
            z.extend(list(_parse(f, _rep(npts - nptloc, (float, 10)))))

        sl = slice(j * npts, (j + 1) * npts)
        map_data[sl, 1] = y
        map_data[sl, 3] = z

    return map_data


def _make_structured_grid(data, method="lagrange3"):
    """Generate a structured grid of unique mach/T4:T2/alt values in the deck"""
    # step size in t4/t2 ratio used in generating the structured grid
    # t2t2_step = 0.5 # original value
    t4t2_step = 0.5
    # step size in mach number used in generating the structured grid
    # mach_step = 0.02 # original value
    mach_step = 0.05

    structured_data = {}

    tt4 = data['thrust'][:, 1]
    tma = data['thrust'][:, 2]
    t4t2s = np.arange(min(tt4), max(tt4) + t4t2_step, t4t2_step)
    machs = np.arange(min(tma), max(tma) + mach_step, mach_step)

    # need t4t2 in first column, mach varies on each row
    pts = np.dstack(np.meshgrid(t4t2s, machs, indexing="ij")).reshape(-1, 2)
    npts = pts.shape[0]

    for field in ["thrust", "fuelflow", "airflow"]:
        map_data = data[field]
        all_alts = map_data[:, 0]
        alts = np.unique(all_alts)

        sizes = (alts.size, t4t2s.size, machs.size)
        vals = np.zeros(np.prod(sizes), dtype=float)
        alt_vec = np.zeros(np.prod(sizes), dtype=float)
        mach_vec = np.zeros(np.prod(sizes), dtype=float)
        t4t2_vec = np.zeros(np.prod(sizes), dtype=float)

        for i, alt in enumerate(alts):
            d = map_data[all_alts == alt]
            t4t2 = np.unique(d[:, 1])
            mach = np.unique(d[:, 2])
            f = d[:, 3].reshape(t4t2.size, mach.size)

            # would explicitly use lagrange3 here to mimic GASP, but some engine
            # decks may not have enough points per dimension
            interp = InterpND(
                method="2D-" + method, points=(t4t2, mach), values=f, extrapolate=True
            )
            sl = slice(i * npts, (i + 1) * npts)
            vals[sl] = interp.interpolate(pts)
            alt_vec[sl] = [alt] * len(pts)
            t4t2_vec[sl] = pts[:, 0]
            mach_vec[sl] = pts[:, 1]

        structured_data[field] = {
            "vals": vals,
            "alts": alt_vec,
            "t4t2s": t4t2_vec,
            "machs": mach_vec,
        }

    return structured_data


def _generate_flight_idle(data, T4T2, ref_sls_airflow, ref_sfn_idle):
    machs = np.unique(data[MACH])
    alts = np.unique(data[ALTITUDE])

    mach_list, alt_list = np.meshgrid(machs, alts)
    mach_list = mach_list.flatten()
    alt_list = alt_list.flatten()

    nn = len(mach_list)

    prob = om.Problem()

    prob.model.add_subsystem(
        Dynamic.Mission.MACH,
        om.IndepVarComp(
            Dynamic.Mission.MACH,
            mach_list,
            units='unitless'),
        promotes=['*'])

    prob.model.add_subsystem(
        Dynamic.Mission.ALTITUDE,
        om.IndepVarComp(
            Dynamic.Mission.ALTITUDE,
            alt_list,
            units='ft'),
        promotes=['*'])

    prob.model.add_subsystem(
        name='atmosphere', subsys=USatm1976Comp(
            num_nodes=nn), promotes_inputs=[
            ('h', Dynamic.Mission.ALTITUDE)], promotes_outputs=[
                ('temp', Dynamic.Mission.TEMPERATURE), ('pres', Dynamic.Mission.STATIC_PRESSURE)])

    prob.model.add_subsystem(
        name='conversion',
        subsys=AtmosCalc(
            num_nodes=nn),
        promotes_inputs=[
            Dynamic.Mission.MACH,
            Dynamic.Mission.TEMPERATURE,
            Dynamic.Mission.STATIC_PRESSURE],
        promotes_outputs=[
            't2',
            'p2'])

    prob.model.add_subsystem(
        name='flight_idle',
        subsys=CalculateIdle(
            num_nodes=nn,
            ref_sfn_idle=ref_sfn_idle,
            ref_sls_airflow=ref_sls_airflow),
        promotes_inputs=[
            't2',
            'p2',
            'pct_corr_airflow_idle',
            'sfc_idle'],
        promotes_outputs=[
            'idle_thrust',
            'idle_fuelflow'])

    prob.setup()

    prob.run_model()

    idle_thrust = prob.get_val('idle_thrust')
    idle_fuelflow = prob.get_val('idle_fuelflow')

    data[MACH] = np.append(data[MACH], mach_list)
    data[ALTITUDE] = np.append(data[ALTITUDE], alt_list)
    data[THRUST] = np.append(data[THRUST], idle_thrust)
    data[FUEL_FLOW] = np.append(data[FUEL_FLOW], idle_fuelflow)
    T4T2 = np.append(T4T2, np.zeros(nn))

    return data, T4T2


_PSLS_PSF = 2116.22  # SLS pressure in psf
_TSLS_DEGR = 518.67  # SLS temperature in deg R


class CalculateIdle(om.ExplicitComponent):
    '''
    Calculates idle conditions of a GASP engine at a specified flight condition
    Vectorized to calculate values for entire flight regime
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

        self.options.declare(
            'ref_sfn_idle',
            1.0,
            desc='Idle thrust-specific fuel consumption, from engine deck')
        self.options.declare(
            'ref_sls_airflow',
            1.0,
            desc='Sea-level static airflow of the reference engine')

    def setup(self):
        nn = self.options["num_nodes"]

        self.add_input(
            "t2", _TSLS_DEGR, units="degR", shape=nn, desc="Engine inlet temperature"
        )
        self.add_input(
            "p2", _PSLS_PSF, units="psf", shape=nn, desc="Engine inlet pressure"
        )
        self.add_input(
            "pct_corr_airflow_idle", 0.5, desc="Percent corrected airflow at idle"
        )
        self.add_input(
            "sfc_idle",
            1.0,
            units="lbm/h/lbf",
            desc="Thrust-specific fuel consumption at idle",
        )

        self.add_output("idle_thrust", units="lbf", shape=nn, desc="Idle thrust")
        # self.add_output(
        #     "idle_airflow", units="lbf/s", shape=nn, desc="Idle corrected airflow"
        # )
        self.add_output("idle_fuelflow", units="lbm/h", shape=nn, desc="Idle fuel flow")

    def compute(self, inputs, outputs):
        (
            t2,
            p2,
            pct_corr_airflow_idle,
            sfc_idle,
        ) = inputs.values()

        ref_sls_airflow = self.options['ref_sls_airflow']
        ref_sfn_idle = self.options['ref_sfn_idle']

        rthet2 = np.sqrt(t2 / _TSLS_DEGR)
        delta2 = p2 / _PSLS_PSF

        airflow_ref = pct_corr_airflow_idle * ref_sls_airflow  # don't un-correct
        thrust_ref = airflow_ref * delta2 / rthet2 * ref_sfn_idle
        fuelflow_ref = thrust_ref * sfc_idle

        outputs["idle_thrust"] = thrust_ref
        # outputs["idle_airflow"] = airflow_ref
        outputs["idle_fuelflow"] = fuelflow_ref


class AtmosCalc(om.ExplicitComponent):
    '''
    Calculates T2 and P2 given static temperature and pressure
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']
        self.add_input(Dynamic.Mission.MACH, val=np.zeros(nn),
                       desc='current Mach number', units='unitless')
        self.add_input(Dynamic.Mission.TEMPERATURE, val=np.zeros(nn),
                       desc='current atmospheric temperature', units='degR')
        self.add_input(
            Dynamic.Mission.STATIC_PRESSURE,
            _PSLS_PSF,
            units="psf",
            shape=nn,
            desc="Ambient static pressure")

        self.add_output(
            "t2",
            units="degR",
            shape=nn,
            desc="Engine inlet total temperature")
        self.add_output("p2", units="psf", shape=nn, desc="Engine inlet total pressure")

    def compute(self, inputs, outputs):
        mach, T, P = inputs.values()

        gamma = 1.4
        t2 = T * (1 + 0.5 * (gamma - 1) * mach**2)
        p2 = P * (t2 / T) ** (gamma / (gamma - 1))

        outputs["t2"] = t2
        outputs["p2"] = p2


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Converts FLOPS- or GASP-formatted '
                                     'engine decks into Aviary csv format.\nFLOPS decks '
                                     'are changed from column-delimited to csv format '
                                     'with added headers.\nGASP decks are reorganized '
                                     'into column based csv. T4 is recovered through '
                                     'calculation. Data points whose T4 exceeds T4max '
                                     'are removed.')
    parser.add_argument('input_file', type=str,
                        help='path to engine deck file to be converted')
    parser.add_argument('output_file', type=str,
                        help='path to file where new converted data will be written')
    parser.add_argument('data_format', type=EngineDeckType, choices=list(EngineDeckType),
                        help='data format used by input_file')

    args = parser.parse_args()

    EngineDeckConverter(**vars(args))



import unittest

import numpy as np
from importlib import import_module
import openmdao.api as om
from openmdao.core.system import System

from aviary.subsystems.subsystem_builder_base import SubsystemBuilderBase
from aviary.utils.aviary_values import AviaryValues


def skipIfMissingDependencies(builder):
    return unittest.skipIf(type(builder) is str, builder)


class TestSubsystemBuilderBase(unittest.TestCase):
    @staticmethod
    def import_builder(path_to_builder: str, base_package='aviary.examples.external_subsystems'):
        '''
        import a subsytem builder

        This is intended to be used with skipIfMissingDependencies
        '''
        try:
            package, method = path_to_builder.rsplit('.', 1)
            package_path, package_name = package.rsplit('.', 1)
            module_path = '.'.join([path_to_builder, package_path]) if package_path \
                else path_to_builder
            module = import_module(package_name, module_path)
            builder = getattr(module, method)
        except ImportError:
            builder = 'Skipping due to missing dependencies'
        except AttributeError:
            builder = method + ' could not be imported from ' + base_package + '.' + package
        return builder

    def setUp(self):
        self.subsystem_builder = SubsystemBuilderBase()
        self.aviary_values = AviaryValues()

    def test_get_states(self):
        states = self.subsystem_builder.get_states()
        self.assertIsInstance(states, dict, msg='get_states should return a dictionary')

        for key, value in states.items():
            self.assertIsInstance(
                value, dict, msg=f"the value for '{key}' should be a dictionary")

            self.assertIsInstance(
                value['rate_source'], str, msg=f"the value for 'rate_source' key in '{key}' should be a string")

    def test_get_linked_variables(self):
        linked_variables = self.subsystem_builder.get_linked_variables()
        self.assertIsInstance(linked_variables, list)
        for var in linked_variables:
            self.assertIsInstance(var, str)

    def test_get_bus_variables(self):
        bus_variables = self.subsystem_builder.get_bus_variables()

        # Check that a dictionary is returned
        self.assertIsInstance(
            bus_variables, dict, 'get_bus_variables should return a dictionary')

        for name, values in bus_variables.items():
            # Check that the bus_variable has the required keys
            self.assertIn('mission_name', values.keys(),
                          f'Bus Variable "{name}" is missing the "mission_name" key')
            self.assertIn('units', values.keys(),
                          f'Bus Variable "{name}" is missing the "units" key')

            # Check that the values of the keys are the expected types
            self.assertIsInstance(
                values['mission_name'], (str, type(None)), f'Bus Variable "{name}"\'s "mission_name" value should be a string or None')
            self.assertIsInstance(
                values['units'], str, f'Bus Variable "{name}"\'s "units" value should be a string')

    def test_build_pre_mission(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()

        pre_mission_sys = self.subsystem_builder.build_pre_mission(
            aviary_inputs=self.aviary_values)

        if pre_mission_sys is not None:
            # Check that pre_mission_sys is an OpenMDAO System
            self.assertIsInstance(
                pre_mission_sys, System, msg="The returned object from `build_pre_mission` is not an OpenMDAO System.")

    def test_build_mission(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()
        # Test that the method returns an OpenMDAO System object
        mission_sys = self.subsystem_builder.build_mission(
            10, aviary_inputs=self.aviary_values)
        if mission_sys is not None:
            self.assertIsInstance(mission_sys, System,
                                  "The method should return an OpenMDAO System object.")

        with self.assertRaises(TypeError, msg='num_nodes argument missing from build_mission().'):
            self.subsystem_builder.build_mission(aviary_inputs=self.aviary_values)

    def test_get_constraints(self):
        constraints = self.subsystem_builder.get_constraints()

        # Check that a dictionary is returned
        self.assertIsInstance(
            constraints, dict, 'get_constraints should return a dictionary')

        for name, values in constraints.items():
            # Check that the constraint has the required keys
            self.assertIn('type', values.keys(),
                          f'Constraint "{name}" is missing the "type" key')

            # Check that the values of the keys are the expected types
            self.assertIsInstance(
                values['type'], str, f'Constraint "{name}"\'s "type" value should be a string')

    def test_get_design_vars(self):
        # Verify that the method returns a dictionary
        design_vars = self.subsystem_builder.get_design_vars()
        self.assertIsInstance(
            design_vars, dict, "get_design_vars() should return a dictionary")

        # Verify that the keys in the dictionary are strings
        for key in design_vars.keys():
            self.assertIsInstance(
                key, str, "The keys in the dictionary returned by get_design_vars() should be strings")

        # Verify that the values in the dictionary are also dictionaries
        for val in design_vars.values():
            self.assertIsInstance(
                val, dict, "The values in the dictionary returned by get_design_vars() should be dictionaries")

        # Verify that the dictionaries have the correct keys
        for val in design_vars.values():
            self.assertIn(
                'units', val, "The dictionaries returned by get_design_vars() should have a 'units' key")
            self.assertIn(
                'lower', val, "The dictionaries returned by get_design_vars() should have a 'lower' key")
            self.assertIn(
                'upper', val, "The dictionaries returned by get_design_vars() should have an 'upper' key")

    def test_get_parameters(self):
        # Verify that the method returns a dictionary
        parameters = self.subsystem_builder.get_parameters()
        self.assertIsInstance(
            parameters, dict, "get_parameters() should return a dictionary")

        # Verify that the keys in the dictionary are strings
        for key in parameters.keys():
            self.assertIsInstance(
                key, str, "The keys in the dictionary returned by get_parameters() should be strings")

        # Verify that the values in the dictionary are also dictionaries
        for val in parameters.values():
            self.assertIsInstance(
                val, dict, "The values in the dictionary returned by get_parameters() should be dictionaries")

        # Verify that the dictionaries have the correct keys
        for val in parameters.values():
            self.assertIn(
                'val', val, "The dictionaries returned by get_parameters() should have a 'val' key")
            self.assertIn(
                'units', val, "The dictionaries returned by get_parameters() should have a 'units' key")

    def test_get_initial_guesses(self):
        initial_guesses = self.subsystem_builder.get_initial_guesses()

        self.assertIsInstance(initial_guesses, dict,
                              msg="get_initial_guesses() should return a dictionary")
        for key, val in initial_guesses.items():
            self.assertIsInstance(
                val, dict, msg=f"The value for '{key}' should be a dictionary")
            self.assertIn(
                'val', val, msg=f"The value for '{key}' dictionary should have a 'val' key")
            self.assertIn(
                'type', val, msg=f"The value for '{key}' dictionary should have a 'type' key")

            guess_val = val['val']
            self.assertIsInstance(guess_val, (float, list, np.ndarray),
                                  msg=f"The value for '{key}' should be a float, list, or array")

    def test_get_mass_names(self):
        # Instantiate a user-defined subsystem builder
        subsystem_builder = self.subsystem_builder

        # Ensure get_mass_names method returns a list
        self.assertIsInstance(subsystem_builder.get_mass_names(), list,
                              "get_mass_names should return a list of mass names")

        # Ensure each mass name is a string
        for mass_name in subsystem_builder.get_mass_names():
            self.assertIsInstance(mass_name, str,
                                  "Each mass name in the list should be a string")

    def test_preprocess_inputs(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()
        inputs = self.subsystem_builder.preprocess_inputs(self.aviary_values)
        self.assertIsInstance(inputs, AviaryValues,
                              "preprocess_inputs did not return an AviaryValues object")

    def test_build_post_mission(self):
        # Perform post-mission operations
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()
        post_mission_sys = self.subsystem_builder.build_post_mission(self.aviary_values)

        if post_mission_sys is not None:
            # Check that post_mission_sys is an OpenMDAO system
            self.assertIsInstance(
                post_mission_sys, System, msg="post_mission_sys is not an OpenMDAO System.")

    def test_define_order(self):
        order = self.subsystem_builder.define_order()
        self.assertIsInstance(order, list, "define_order should return a list")

        for subsystem_name in order:
            self.assertIsInstance(
                subsystem_name, str, "Each subsystem name in the list should be a string")

    def test_get_outputs(self):
        outputs = self.subsystem_builder.get_outputs()
        self.assertIsInstance(outputs, list, "get_outputs should return a list")

        for output_name in outputs:
            self.assertIsInstance(
                output_name, str, "Each output name in the list should be a string")

    def test_check_state_variables(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()

        states = self.subsystem_builder.get_states()

        mission_sys = self.subsystem_builder.build_mission(
            num_nodes=5, aviary_inputs=self.aviary_values)

        if mission_sys is None:
            return

        group = om.Group()
        group.add_subsystem('mission', mission_sys, promotes=['*'])
        prob = om.Problem(group)
        prob.setup()
        prob.final_setup()

        inputs = prob.model.list_inputs(out_stream=None, prom_name=True)
        outputs = prob.model.list_outputs(out_stream=None, prom_name=True)

        for key, value in states.items():
            if mission_sys is not None:
                # Check state variable existence
                state_var_exists = any(key == input[1]['prom_name'] for input in inputs)
                self.assertTrue(state_var_exists,
                                f"State variable '{key}' not found in the model.")

                # Check rate_source variable existence
                if 'rate_source' in value:
                    rate_source = value['rate_source']
                    rate_source_exists = any(
                        rate_source in output[0] for output in outputs)
                    self.assertTrue(
                        rate_source_exists, f"Rate source variable '{rate_source}' for state '{key}' not found in the model.")

    def test_check_pre_mission(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()

        pre_mission_sys = self.subsystem_builder.build_pre_mission(self.aviary_values)

        if pre_mission_sys is None:
            return

        group = om.Group()
        group.add_subsystem('pre_mission_sys', pre_mission_sys)
        prob = om.Problem(group)
        prob.setup()
        prob.final_setup()

        self.assertIsInstance(pre_mission_sys, System,
                              "The method should return an OpenMDAO System object.")

        mass_names = self.subsystem_builder.get_mass_names()
        outputs = pre_mission_sys.list_outputs(out_stream=None, prom_name=True)

        for name in mass_names:
            mass_var_exists = any(name == output_[1]['prom_name'] for output_ in outputs)
            self.assertTrue(
                mass_var_exists, f"Mass variable '{name}' not found in the pre-mission model.")

    def test_check_parameters(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()

        parameters = self.subsystem_builder.get_parameters()

        mission_sys = self.subsystem_builder.build_mission(
            num_nodes=5, aviary_inputs=self.aviary_values)

        if mission_sys is None:
            return

        group = om.Group()
        group.add_subsystem('mission', mission_sys, promotes=['*'])
        prob = om.Problem(group)
        prob.setup()
        prob.final_setup()

        inputs = prob.model.list_inputs(out_stream=None, prom_name=True)

        for key, value in parameters.items():
            # Check parameter existence
            param_exists = any(key == input_[1]['prom_name'] for input_ in inputs)
            self.assertTrue(param_exists, f"Parameter '{key}' not found in the model.")

    def test_check_constraints(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()

        constraints = self.subsystem_builder.get_constraints()

        mission_sys = self.subsystem_builder.build_mission(
            num_nodes=5, aviary_inputs=self.aviary_values)

        if mission_sys is None:
            return

        group = om.Group()
        group.add_subsystem('mission', mission_sys, promotes=['*'])
        prob = om.Problem(group)
        prob.setup()
        prob.final_setup()

        inputs = prob.model.list_inputs(out_stream=None, prom_name=True)
        outputs = prob.model.list_outputs(out_stream=None, prom_name=True)

        for key, value in constraints.items():
            # Check constraint existence
            constraint_exists = (any(key == output[1]['prom_name'] for output in outputs) or any(
                key == input[1]['prom_name'] for input in inputs))
            self.assertTrue(constraint_exists,
                            f"Constraint '{key}' not found in the model.")

    def test_check_design_variables(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()

        design_vars = self.subsystem_builder.get_design_vars()

        pre_mission_sys = self.subsystem_builder.build_pre_mission(
            aviary_inputs=self.aviary_values)

        if pre_mission_sys is None:
            return

        group = om.Group()
        group.add_subsystem('pre_mission', pre_mission_sys, promotes=['*'])
        prob = om.Problem(group)
        prob.setup()
        prob.final_setup()

        inputs = prob.model.list_inputs(out_stream=None, prom_name=True)

        for key, value in design_vars.items():
            # Check design variable existence
            design_var_exists = any(key == input_[1]['prom_name'] for input_ in inputs)
            self.assertTrue(design_var_exists,
                            f"Design variable '{key}' not found in the model.")

    def test_check_initial_guesses(self):
        if not hasattr(self, 'aviary_values'):
            self.aviary_values = AviaryValues()

        initial_guesses = self.subsystem_builder.get_initial_guesses()

        mission_sys = self.subsystem_builder.build_mission(
            num_nodes=5, aviary_inputs=self.aviary_values)

        if mission_sys is None:
            return

        group = om.Group()
        group.add_subsystem('mission', mission_sys, promotes=['*'])
        prob = om.Problem(group)
        prob.setup()
        prob.final_setup()

        inputs = prob.model.list_inputs(out_stream=None, prom_name=True)

        for key, value in initial_guesses.items():
            # Check initial guess existence
            initial_guess_exists = any(key == input[1]['prom_name'] for input in inputs)
            self.assertTrue(initial_guess_exists,
                            f"Initial guess '{key}' not found in the model.")


if __name__ == '__main__':
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


# NOTE default values for avg diam & avg length if not defined by user:
#      Aircraft.Nacelle.AVG_LENGTH = 0.07 * sqrt(Aircraft.ENGINE.SCALED_SLS_THRUST)
#      Aircraft.Nacelle.AVG_DIAMETER = 0.04 * sqrt(Aircraft.ENGINE.SCALED_SLS_THRUST)
class Nacelles(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        count = len(self.options['aviary_options'].get_val(Aircraft.Engine.NUM_ENGINES))
        add_aviary_input(self, Aircraft.Nacelle.AVG_DIAMETER, val=np.zeros(count))
        add_aviary_input(self, Aircraft.Nacelle.AVG_LENGTH, val=np.zeros(count))
        add_aviary_input(self, Aircraft.Nacelle.WETTED_AREA_SCALER, val=np.zeros(count))

        add_aviary_output(self, Aircraft.Nacelle.TOTAL_WETTED_AREA, 0.0)
        add_aviary_output(self, Aircraft.Nacelle.WETTED_AREA, val=np.zeros(count))

    def setup_partials(self):
        self.declare_partials(
            Aircraft.Nacelle.TOTAL_WETTED_AREA,
            [
                Aircraft.Nacelle.AVG_DIAMETER, Aircraft.Nacelle.AVG_LENGTH,
                Aircraft.Nacelle.WETTED_AREA_SCALER,
            ]
        )

        self.declare_partials(
            Aircraft.Nacelle.WETTED_AREA,
            [
                Aircraft.Nacelle.AVG_DIAMETER, Aircraft.Nacelle.AVG_LENGTH,
                Aircraft.Nacelle.WETTED_AREA_SCALER,
            ]
        )

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_engines = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)
        count = len(num_engines)

        avg_diam = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        avg_length = inputs[Aircraft.Nacelle.AVG_LENGTH]
        scaler = inputs[Aircraft.Nacelle.WETTED_AREA_SCALER]

        wetted_area = np.zeros(count, dtype=avg_diam.dtype)

        calc_idx = np.where(num_engines >= 1)
        wetted_area[calc_idx] = scaler[calc_idx] * 2.8 * \
            avg_diam[calc_idx] * avg_length[calc_idx]

        outputs[Aircraft.Nacelle.WETTED_AREA] = wetted_area

        total_wetted_area = sum(num_engines * wetted_area)

        outputs[Aircraft.Nacelle.TOTAL_WETTED_AREA] = total_wetted_area

    def compute_partials(self, inputs, J, discrete_inputs=None):
        aviary_options: AviaryValues = self.options['aviary_options']
        num_engines = aviary_options.get_val(Aircraft.Engine.NUM_ENGINES)

        avg_diam = inputs[Aircraft.Nacelle.AVG_DIAMETER]
        avg_length = inputs[Aircraft.Nacelle.AVG_LENGTH]
        scaler = inputs[Aircraft.Nacelle.WETTED_AREA_SCALER]

        deriv_area_len = np.zeros(len(num_engines), dtype=avg_diam.dtype)
        deriv_area_diam = np.zeros(len(num_engines), dtype=avg_diam.dtype)
        deriv_area_scaler = np.zeros(len(num_engines), dtype=avg_diam.dtype)
        deriv_total_len = np.zeros(len(num_engines), dtype=avg_diam.dtype)
        deriv_total_diam = np.zeros(len(num_engines), dtype=avg_diam.dtype)
        deriv_total_scaler = np.zeros(len(num_engines), dtype=avg_diam.dtype)

        area_to_length = 2.8 * avg_diam
        area_to_diam = 2.8 * avg_length
        area_to_scaler = 2.8 * avg_diam * avg_length

        calc_idx = np.where(num_engines >= 1)
        deriv_area_len[calc_idx] = scaler[calc_idx] * area_to_length[calc_idx]
        deriv_area_diam[calc_idx] = scaler[calc_idx] * area_to_diam[calc_idx]
        deriv_area_scaler[calc_idx] = area_to_scaler[calc_idx]
        deriv_total_len[calc_idx] = scaler[calc_idx] * \
            num_engines[calc_idx] * area_to_length[calc_idx]
        deriv_total_diam[calc_idx] = scaler[calc_idx] * \
            num_engines[calc_idx] * area_to_diam[calc_idx]
        deriv_total_scaler[calc_idx] = num_engines[calc_idx] * area_to_scaler[calc_idx]

        J[Aircraft.Nacelle.WETTED_AREA, Aircraft.Nacelle.AVG_LENGTH] = deriv_area_len

        J[Aircraft.Nacelle.WETTED_AREA, Aircraft.Nacelle.AVG_DIAMETER] = deriv_area_diam

        J[Aircraft.Nacelle.WETTED_AREA, Aircraft.Nacelle.WETTED_AREA_SCALER] = deriv_area_scaler

        J[Aircraft.Nacelle.TOTAL_WETTED_AREA, Aircraft.Nacelle.AVG_LENGTH] = deriv_total_len

        J[Aircraft.Nacelle.TOTAL_WETTED_AREA, Aircraft.Nacelle.AVG_DIAMETER] = deriv_total_diam

        J[Aircraft.Nacelle.TOTAL_WETTED_AREA,
            Aircraft.Nacelle.WETTED_AREA_SCALER] = deriv_total_scaler



import os
from pathlib import Path
import argparse
import pkg_resources
import shutil


def aviary_resource(resource_name: str) -> str:
    return pkg_resources.resource_filename("aviary", resource_name)


def get_model(file_name: str, verbose=False) -> Path:
    '''
    This function attempts to find the path to a file or folder in aviary/models
    If the path cannot be found in any of the locations, a FileNotFoundError is raised.

    Parameters
    ----------
    path : str or Path
        The input path, either as a string or a Path object.

    Returns
    -------
    aviary_path
        The absolute path to the file.

    Raises
    ------
    FileNotFoundError
        If the path is not found.
    '''

    # Get the path to Aviary's models
    path = Path('models', file_name)
    aviary_path = Path(aviary_resource(str(path)))

    # If the file name was provided without a path, check in the subfolders
    if not aviary_path.exists():
        sub_dirs = [x[0] for x in os.walk(aviary_resource('models'))]
        for sub_dir in sub_dirs:
            temp_path = Path(sub_dir, file_name)
            if temp_path.exists():
                # only return the first matching file
                aviary_path = temp_path
                continue

    # If the path still doesn't exist, raise an error.
    if not aviary_path.exists():
        raise FileNotFoundError(
            f"File or Folder not found in Aviary's hangar"
        )
    if verbose:
        print('found', aviary_path, '\n')
    return aviary_path


def save_file(aviary_path: Path, outdir: Path, verbose=False) -> Path:
    '''
    Saves the file or folder specified into the output directory, creating directories as needed.
    '''
    outdir.mkdir(parents=True, exist_ok=True)
    if aviary_path.is_dir():
        if verbose:
            print(aviary_path, 'is a directory, getting all files')
        outdir = outdir.joinpath(aviary_path.stem)
        outdir.mkdir(exist_ok=True)
        for file in next(os.walk(aviary_path))[-1]:
            if verbose:
                print('copying', str(aviary_path / file), 'to', str(outdir / file))
            shutil.copy2(aviary_path / file, outdir)
    else:
        if verbose:
            print('copying', str(aviary_path), 'to', str(outdir / aviary_path.name))
        shutil.copy2(aviary_path, outdir)
    return outdir


def _setup_hangar_parser(parser: argparse.ArgumentParser):
    def_outdir = os.path.join(os.getcwd(), "aviary_models")
    parser.add_argument(
        'input_decks', metavar='indecks', type=str, nargs='+', help='Name of file or folder to download from Aviary/models'
    )
    parser.add_argument(
        "-o", "--outdir", default=def_outdir, help="Directory to write outputs. Defaults to aviary_models in the current directory."
    )
    parser.add_argument(
        "-v", "--verbose",
        action="store_true",
        help="Enable verbose outputs",
    )


def _exec_hangar(args, user_args):
    # check if args.input_deck is a list, if so, use the first element
    input_decks = []
    for input_deck in args.input_decks:
        input_decks.append(get_model(input_deck, args.verbose))

    for input_deck in input_decks:
        save_file(input_deck, Path(args.outdir), args.verbose)



from aviary.variable_info.variables import Aircraft as av_Aircraft
from aviary.variable_info.variables import Mission as av_Mission

AviaryAircraft = av_Aircraft
AviaryMission = av_Mission


# ---------------------------
# Aircraft data hierarchy
# ---------------------------

class Aircraft(AviaryAircraft):

    # cell = single cell, battery = one case plus multiple cells

    class Battery:
        CURRENT_MAX = "aircraft:battery:current_max"
        EFFICIENCY = "aircraft:battery:efficiency"
        ENERGY_REQUIRED = "aircraft:battery:energy_required"
        HEAT_CAPACITY = "aircraft:battery:heat_capacity"
        MASS = "aircraft:battery:mass"
        N_PARALLEL = "aircraft:battery:n_parallel"
        N_SERIES = "aircraft:battery:n_series"
        VOLTAGE = "aircraft:battery:voltage"
        VOLUME = "aircraft:battery:volume"

        class Case:
            HEAT_CAPACITY = "aircraft:battery:case:heat_capacity"
            WEIGHT_FRAC = "aircraft:battery:case:weight_frac"

        class Cell:
            DISCHARGE_RATE = "aircraft:battery:cell:discharge_rate"
            ENERGY_CAPACITY_MAX = "aircraft:battery:cell:energy_capacity_max"
            HEAT_CAPACITY = "aircraft:battery:cell:heat_capacity"
            MASS = "aircraft:battery:cell:mass"
            VOLTAGE_LOW = "aircraft:battery:cell:voltage_low"
            VOLUME = "aircraft:battery:cell:volume"
            TYPE = "aircraft:battery:cell:type"

# ---------------------------
# Mission data hierarchy
# ---------------------------


class Mission(AviaryMission):

    class Battery:
        CURRENT = "mission:battery:current"
        HEAT_OUT = "mission:battery:heat_out"
        STATE_OF_CHARGE = "mission:battery:state_of_charge"
        STATE_OF_CHARGE_RATE = "mission:battery:state_of_charge_rate"
        TEMPERATURE = "mission:battery:temperature"
        VOLTAGE = "mission:battery:voltage"
        VOLTAGE_THEVENIN = "mission:battery:voltage_thevenin"
        VOLTAGE_THEVENIN_RATE = "mission:battery:voltage_thevenin_rate"



import numpy as np
import openmdao.api as om
from aviary.mission.gasp_based.ode.time_integration_base_classes import SimuPyProblem, SGMTrajBase
from aviary.mission.gasp_based.phases.time_integration_phases import SGMGroundroll, SGMRotation, SGMAscentCombined, SGMAccel, SGMClimb, SGMCruise, SGMDescent

from aviary.variable_info.enums import SpeedType
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Mission, Dynamic

DEBUG = 0


class TimeIntegrationTrajBase(SGMTrajBase):
    def initialize(self):
        super().initialize()
        self.options.declare("cruise_mach", default=0.8)
        self.options.declare("ode_args", types=dict, default=dict())
        self.options.declare("ode_args_pyc", types=dict, default=dict())
        self.options.declare("pyc_phases", default=list())


class FlexibleTraj(TimeIntegrationTrajBase):
    '''
    A traj that can be given a list of phases to build flexible trajectories.
    This is useful for simulating/testing phases one at a time as well as
    allowing users to quickly redefine the phase order during the tracjectory.
    '''

    def initialize(self):
        super().initialize()
        self.options.declare('Phases', default=None)
        self.options.declare('promote_all_auto_ivc', default=False)
        self.options.declare('traj_final_state_output', default=None)
        self.options.declare('traj_promote_final_output', default=None)
        self.options.declare('traj_promote_initial_input', default=None)
        self.options.declare('traj_initial_state_input', default=None)
        self.options.declare('traj_event_trigger_input', default=None)

    def setup(self):

        ODEs = []
        for phase_name, phase_info in self.options['Phases'].items():
            next_phase = phase_info['ode']
            next_phase.phase_name = phase_name
            ODEs.append(next_phase)

        self.setup_params(
            ODEs=ODEs,
            promote_all_auto_ivc=self.options['promote_all_auto_ivc'],
            traj_final_state_output=self.options['traj_final_state_output'],
            traj_promote_final_output=self.options['traj_promote_final_output'],

            traj_promote_initial_input=self.options['traj_promote_initial_input'],
            traj_initial_state_input=self.options['traj_initial_state_input'],
            traj_event_trigger_input=self.options['traj_event_trigger_input'],
        )
        self.declare_partials(["*"], ["*"],)

    def compute(self, inputs, outputs):
        self.compute_params(inputs)

        for phase in self.ODEs:
            phase_name = phase.phase_name
            vals_to_set = self.options['Phases'][phase_name]['vals_to_set']
            if vals_to_set:
                for name, data in vals_to_set.items():
                    if name.startswith('attr:'):
                        setattr(phase, name.replace('attr:', ''), inputs[data['val']])
                    elif name.startswith('rotation.'):
                        phase.rotation.set_val(name.replace(
                            'rotation.', ''), data['val'], units=data['units'])
                    else:
                        phase.set_val(name, data['val'], units=data['units'])

        ode_index = 0
        sim_gen = self.compute_traj_loop(self.ODEs[0], inputs, outputs)
        print('*'*40)
        print('Starting: '+self.ODEs[ode_index].phase_name)
        for current_problem, sim_result in sim_gen:
            t_final = sim_result.t[-1]
            x_final = sim_result.x[-1, :]
            if type(current_problem) is SGMGroundroll:
                t_start_rotation = t_final

            ode_index += 1
            try:
                next_problem = self.ODEs[ode_index]
            except IndexError:
                next_problem = None

            print('Finished: '+current_problem.phase_name)
            if next_problem is not None:
                if type(current_problem) is SGMGroundroll:
                    next_problem.prob.set_val("start_rotation", t_start_rotation)
                elif type(current_problem) is SGMRotation:
                    next_problem.rotation.set_val("start_rotation", t_start_rotation)
                print('Starting: '+next_problem.phase_name)
                sim_gen.send(next_problem)
            else:
                print('Reached the end of the Trajectory!')
                sim_gen.close()
                break

        print('t_final', t_final)
        print('x_final', x_final)
        print(self.ODEs[-1].states)

# class SGMTraj1(TimeIntegrationTrajBase):
#     '''
#     This combines the phases from brake release to landing
#     '''

#     def initialize(self):
#         super().initialize()

#     def setup(self):
#         """
#         API requirements:
#             pass ODE's,
#             next_problem = f(current_problem, current_result)
#             initial_state/time/etc
#             next_state from last state/output/event information

#             pass in terminal and integrand output functions with derivatives (components)
#             -- anything special for final state, final time?
#             declare initial state(s) as parameters to take derivative wrt
#             assume all other inputs are parameters for deriv?
#         """

#         # distinction is really "meta" inputs to the trajectory itself and ODE
#         # parameters that should get automatically passed (previously `passed_inputs`)
#         # may also need to just have a specific call out for initial condition
#         # parameters?

#         # the "meta" inputs are really initial condition and/or event termination
#         # paramters, like initial states or state triggers
#         # need an API for that
#         # I guess there are similarly at least two categories of outputs, terminal and
#         # integral terms. or just say ode must provide it as a state so theyre all meyer
#         # terms

#         # actual ODE setup
#         groundroll = SGMGroundroll(
#             ode_args=(self.options["ode_args_pyc"] if "groundroll" in
#                       self.options["pyc_phases"] else self.options["ode_args"])
#         )
#         rotation = SGMRotation(
#             ode_args=(self.options["ode_args_pyc"] if "groundroll" in
#                       self.options["pyc_phases"] else self.options["ode_args"])
#         )
#         ascent = SGMAscentCombined(
#             ode_args=(self.options["ode_args_pyc"] if "ascent" in
#                       self.options["pyc_phases"] else self.options["ode_args"])
#         )
#         accel = SGMAccel(
#             ode_args=(self.options["ode_args_pyc"] if "accel" in
#                       self.options["pyc_phases"] else self.options["ode_args"])
#         )
#         climb1 = SGMClimb(
#             input_speed_type=SpeedType.EAS,
#             input_speed_units="kn",
#             ode_args=(self.options["ode_args_pyc"] if "climb" in
#                       self.options["pyc_phases"] else self.options["ode_args"])
#         )

#         climb2 = SGMClimb(
#             input_speed_type=SpeedType.EAS,
#             input_speed_units="kn",
#             ode_args=(self.options["ode_args_pyc"] if "climb" in
#                       self.options["pyc_phases"] else self.options["ode_args"])
#         )
#         climb3 = SGMClimb(
#             input_speed_type=SpeedType.MACH,
#             input_speed_units="unitless",
#             ode_args=(self.options["ode_args_pyc"] if "climb" in
#                       self.options["pyc_phases"] else self.options["ode_args"])
#         )
#         self.setup_params(
#             ODEs=[
#                 groundroll, rotation, ascent, accel, climb1, climb2, climb3
#             ],
#             traj_final_state_output=[Dynamic.Mission.MASS, Dynamic.Mission.DISTANCE,],
#             traj_promote_final_output=[Dynamic.Mission.ALTITUDE_RATE,
#                                        Dynamic.Mission.FLIGHT_PATH_ANGLE, "TAS"],

#             traj_initial_state_input=[Dynamic.Mission.MASS],
#             traj_event_trigger_input=[
#                 # specify ODE, output_name, with units that SimuPyProblem expects
#                 # assume event function is of form ODE.output_name - value
#                 # third key is event_idx associated with input
#                 (groundroll, "TAS", 0,),
#                 (climb3, Dynamic.Mission.ALTITUDE, 0,),
#             ],
#         )

#     def compute(self, inputs, outputs):
#         self.compute_params(inputs)

#         # ODE setup
#         (groundroll, rotation, ascent, accel, climb1, climb2, climb3) = self.ODEs

#         cruise_alt = inputs["SGMClimb_"+Dynamic.Mission.ALTITUDE+"_trigger"]
#         cruise_mach = self.options["cruise_mach"]

#         groundroll.VR_value = inputs["SGMGroundroll_TAS_trigger"]

#         climb1.set_val("alt_trigger", 10_000, units="ft")
#         climb2.set_val("alt_trigger", cruise_alt, units="ft")
#         climb3.set_val("alt_trigger", cruise_alt, units="ft")

#         speed_string = {SpeedType.EAS: 'EAS',
#                         SpeedType.TAS: 'TAS', SpeedType.MACH: 'mach'}
#         climb1.set_val(speed_string[climb1.input_speed_type],
#                        250., units=climb1.input_speed_units)
#         climb2.set_val(speed_string[climb2.input_speed_type],
#                        270., units=climb2.input_speed_units)
#         climb3.set_val(speed_string[climb3.input_speed_type], cruise_mach,
#                        units=climb3.input_speed_units)

#         climb1.set_val("speed_trigger", cruise_mach, units=None)
#         climb2.set_val("speed_trigger", cruise_mach, units=None)
#         climb3.set_val("speed_trigger", 0.0, units=None)

#         for t_var in ["t_init_gear", "t_init_flaps"]:
#             ascent.set_val(t_var, 10_000.)
#         ascent.rotation.set_val("start_rotation", 10_000.)

#         sim_gen = self.compute_traj_loop(groundroll, inputs, outputs)

#         for current_problem, sim_result in sim_gen:

#             t = sim_result.t[-1]
#             x = sim_result.x[-1, :]
#             event_idx = np.argmin(np.abs(sim_result.e[-1, :]))

#             # trajectory-specific phase switching
#             if current_problem is groundroll:
#                 if DEBUG:
#                     print("starting rotation")
#                 rotation.prob.set_val("start_rotation", t)
#                 ascent.rotation.set_val("start_rotation", t)
#                 next_problem = rotation
#             elif current_problem is rotation:
#                 if DEBUG:
#                     print("starting ascent")
#                 next_problem = ascent
#             elif current_problem is ascent:
#                 next_problem = accel
#                 if DEBUG:
#                     print("starting accel")
#             elif current_problem is accel:
#                 if DEBUG:
#                     print("climb1")
#                 next_problem = climb1
#             elif current_problem is climb1:
#                 if DEBUG:
#                     print("climb2")
#                 if event_idx != 0:
#                     if DEBUG:
#                         print("expected to hit alt trigger for climb1")
#                 next_problem = climb2
#             elif current_problem is climb2:
#                 if DEBUG:
#                     print("climb3")
#                 if event_idx != 1:
#                     if DEBUG:
#                         print("expected to hit speed trigger for climb2")
#                 next_problem = climb3
#             elif current_problem is climb3:
#                 if DEBUG:
#                     print("climb ending")
#                 if event_idx != 0:
#                     if DEBUG:
#                         print("expected to hit alt trigger for climb3")
#                 next_problem = None
#             else:
#                 if DEBUG:
#                     print("unexpected termination")
#                 next_problem = None

#             if next_problem is not None:
#                 sim_gen.send(next_problem)
#             else:
#                 sim_gen.close()
#                 break


# class SGMTraj2(TimeIntegrationTrajBase):
#     '''
#     This combines the phases from end of cruise to landing
#     '''

#     def setup(self):
#         """
#         API requirements:
#             pass ODE's,
#             next_problem = f(current_problem, current_result)
#             initial_state/time/etc
#             next_state from last state/output/event information

#             pass in terminal and integrand output functions with derivatives (components)
#             -- anything special for final state, final time?
#             declare initial state(s) as parameters to take derivative wrt
#             assume all other inputs are parameters for deriv?
#         """

#         desc1 = SGMDescent(
#             input_speed_type=SpeedType.MACH,
#             input_speed_units="unitless",
#             speed_trigger_units='kn',
#             ode_args=self.options["ode_args"],
#         )

#         desc2 = SGMDescent(
#             input_speed_type=SpeedType.EAS,
#             input_speed_units="kn",
#             speed_trigger_units='kn',
#             ode_args=self.options["ode_args"],
#         )
#         desc3 = SGMDescent(
#             input_speed_type=SpeedType.EAS,
#             input_speed_units="kn",
#             speed_trigger_units='kn',
#             ode_args=self.options["ode_args"],
#         )

#         self.setup_params(
#             ODEs=[desc1, desc2, desc3],
#             traj_final_state_output=[Dynamic.Mission.MASS, Dynamic.Mission.DISTANCE],
#             traj_initial_state_input=[Dynamic.Mission.MASS,
#                                       Dynamic.Mission.DISTANCE, Dynamic.Mission.ALTITUDE],
#         )
#         self.declare_partials(["*"], ["*"],)

#     def compute(self, inputs, outputs):

#         self.compute_params(inputs)

#         # ODE setup
#         desc1, desc2, desc3 = self.ODEs

#         cruise_mach = self.options["cruise_mach"]

#         desc1.set_val("alt_trigger", 10_000, units="ft")
#         desc2.set_val("alt_trigger", 10_000, units="ft")
#         desc3.set_val("alt_trigger", 1_000, units="ft")

#         speed_string = {SpeedType.EAS: 'EAS',
#                         SpeedType.TAS: 'TAS', SpeedType.MACH: 'mach'}
#         desc1.set_val(speed_string[desc1.input_speed_type], cruise_mach,
#                       units=desc1.input_speed_units)
#         desc2.set_val(speed_string[desc2.input_speed_type],
#                       350., units=desc2.input_speed_units)
#         desc3.set_val(speed_string[desc3.input_speed_type],
#                       250., units=desc3.input_speed_units)

#         desc1.set_val("speed_trigger", 350.0, units=desc1.speed_trigger_units)
#         desc2.set_val("speed_trigger", 0.0, units=desc2.speed_trigger_units)
#         desc3.set_val("speed_trigger", 0.0, units=desc3.speed_trigger_units)

#         # main loop
#         sim_gen = self.compute_traj_loop(desc1, inputs, outputs)

#         for current_problem, sim_result in sim_gen:
#             t = sim_result.t[-1]
#             x = sim_result.x[-1, :]

#             event_idx = np.argmin(np.abs(sim_result.e[-1, :]))

#             # trajectory-specific phase switching
#             if current_problem is desc1:
#                 if DEBUG:
#                     print("desc2")
#                 if event_idx != 1:
#                     if DEBUG:
#                         print("expected to hit speed trigger for desc1")
#                 sim_gen.send(desc2)
#             elif current_problem is desc2:
#                 if DEBUG:
#                     print("desc3")
#                 if event_idx != 0:
#                     if DEBUG:
#                         print("expected to hit alt trigger for desc2")
#                 sim_gen.send(desc3)
#             elif current_problem is desc3:
#                 if DEBUG:
#                     print("desc ending")
#                 if event_idx != 0:
#                     if DEBUG:
#                         print("expected to hit alt trigger for desc3")
#                 sim_gen.close()
#             else:
#                 if DEBUG:
#                     print("unexpected termination")
#                 sim_gen.close()



import enum
import numpy as np

import openmdao.api as om

from aviary.mission.gasp_based.flight_conditions import FlightConditions
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.enums import AnalysisScheme, AlphaModes, SpeedType
from aviary.variable_info.variables import Aircraft, Mission, Dynamic
from aviary.mission.ode.specific_energy_rate import SpecificEnergyRate
from aviary.mission.ode.altitude_rate import AltitudeRate


class BaseODE(om.Group):
    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)
        self.options.declare(
            "analysis_scheme",
            default=AnalysisScheme.COLLOCATION,
            types=enum.Enum,
            desc="The analysis method that will be used to close the trajectory; for example collocation or time integration",
        )

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

        # TODO finish description
        self.options.declare(
            'core_subsystems',
            desc='list of core subsystems'
        )

        self.options.declare(
            'subsystem_options', types=dict, default={},
            desc='dictionary of parameters to be passed to the subsystem builders'
        )

    def AddAlphaControl(
        self,
        alpha_group=None,
        alpha_mode: AlphaModes = AlphaModes.DEFAULT,
        num_nodes=1,
        target_load_factor=1.1,
        target_tas_rate=0,
        # target_alt_rate=0,
        atol=1e-7,
        rtol=1e-7,
        add_default_solver=True,
        print_level=0,
    ):
        '''
        This is used when angle of attack in an ODE needs to be controlled directly.
        '''

        if not alpha_group:
            alpha_group = self
        nn = num_nodes

        if alpha_mode is AlphaModes.ROTATION:
            alpha_comp = om.ExecComp(
                'alpha=rotation_rate*(t_curr-start_rotation)+alpha_init',
                alpha=dict(val=0., units='deg'),
                rotation_rate=dict(val=10.0/3.0, units='deg/s'),
                t_curr=dict(val=0., units='s'),
                start_rotation=dict(val=0., units='s'),
                alpha_init=dict(val=0., units='deg'),
            )
            alpha_comp_inputs = ["rotation_rate", "t_curr", "start_rotation",
                                 ("alpha_init", Aircraft.Wing.INCIDENCE)]

        elif alpha_mode is AlphaModes.LOAD_FACTOR:
            alpha_comp = om.BalanceComp(
                name="alpha",
                val=np.full(nn, 10),  # initial guess
                units="deg",
                eq_units="unitless",
                lhs_name="load_factor",
                rhs_val=target_load_factor,
                upper=25.,
                lower=-2.,
            )
            alpha_comp_inputs = ["load_factor"]

        elif alpha_mode is AlphaModes.FUSELAGE_PITCH:
            alpha_comp = om.ExecComp(
                'alpha=max_fus_angle-gamma+i_wing',
                alpha=dict(val=0., units='deg'),
                max_fus_angle=dict(val=0., units='deg'),
                gamma=dict(val=0., units='deg'),
                i_wing=dict(val=0., units='deg'),
            )
            alpha_comp_inputs = [("max_fus_angle", Aircraft.Design.MAX_FUSELAGE_PITCH_ANGLE),
                                 ("gamma", Dynamic.Mission.FLIGHT_PATH_ANGLE),
                                 ("i_wing", Aircraft.Wing.INCIDENCE)]

        elif alpha_mode is AlphaModes.DECELERATION:
            alpha_comp = om.BalanceComp(
                name="alpha",
                val=np.full(nn, 10),  # initial guess
                units="deg",
                lhs_name=Dynamic.Mission.VELOCITY_RATE,
                rhs_name='target_tas_rate',
                rhs_val=target_tas_rate,
                eq_units="kn/s",
                upper=25.,
                lower=-2.,
            )
            alpha_comp_inputs = [Dynamic.Mission.VELOCITY_RATE]

        elif alpha_mode is AlphaModes.REQUIRED_LIFT:
            alpha_comp = om.BalanceComp(
                name="alpha",
                val=8.0 * np.ones(nn),
                units="deg",
                rhs_name="required_lift",
                lhs_name=Dynamic.Mission.LIFT,
                eq_units="lbf",
                upper=12.0,
                lower=-2,
            )
            alpha_comp_inputs = ["required_lift", Dynamic.Mission.LIFT]

        # Future controller modes
        # elif alpha_mode is AlphaModes.ALTITUDE_RATE:
        #     alpha_comp = om.BalanceComp(
        #         name="alpha",
        #         val=np.full(nn, 1),
        #         units="deg",
        #         lhs_name=Dynamic.Mission.ALTITUDE_RATE,
        #         rhs_name='target_alt_rate',
        #         rhs_val=target_alt_rate,
        #         upper=12.0,
        #         lower=-2,
        #     )
        #     alpha_comp_inputs = [Dynamic.Mission.ALTITUDE_RATE]

        # elif alpha_mode is AlphaModes.CONSTANT_ALTITUDE:
        #     alpha_comp = om.BalanceComp(
        #         name="alpha",
        #         val=np.full(nn, 1),
        #         units="deg",
        #         lhs_name=Dynamic.Mission.ALTITUDE,
        #         rhs_name='target_alt',
        #         rhs_val=37500,
        #         upper=12.0,
        #         lower=-2,
        #     )
        #     alpha_comp_inputs = [Dynamic.Mission.ALTITUDE]

        if alpha_mode is not AlphaModes.DEFAULT:
            alpha_group.add_subsystem("alpha_comp",
                                      alpha_comp,
                                      promotes_inputs=alpha_comp_inputs,
                                      promotes_outputs=["alpha"],
                                      )

            if add_default_solver and alpha_mode not in (AlphaModes.ROTATION,):
                alpha_group.nonlinear_solver = om.NewtonSolver()
                alpha_group.nonlinear_solver.options["solve_subsystems"] = True
                alpha_group.nonlinear_solver.options["iprint"] = print_level
                alpha_group.nonlinear_solver.options["atol"] = atol
                alpha_group.nonlinear_solver.options["rtol"] = rtol
                alpha_group.nonlinear_solver.linesearch = om.BoundsEnforceLS()
                alpha_group.linear_solver = om.DirectSolver(assemble_jac=True)

    def AddThrottleControl(
        self,
        prop_group=om.Group(),
        num_nodes=1,
        atol=1e-12,
        rtol=1e-12,
        add_default_solver=True,
        print_level=0,
    ):
        '''
        This is used when throttle in an ODE needs to be controlled directly.
        '''

        nn = num_nodes

        thrust_bal = om.BalanceComp(
            name=Dynamic.Mission.THROTTLE,
            val=np.ones(nn),
            upper=1.0,
            lower=0.0,
            units='unitless',
            lhs_name=Dynamic.Mission.THRUST_TOTAL,
            rhs_name="required_thrust",
            eq_units="lbf",
        )
        prop_group.add_subsystem("thrust_balance",
                                 thrust_bal,
                                 promotes_inputs=[
                                     Dynamic.Mission.THRUST_TOTAL, 'required_thrust'],
                                 promotes_outputs=[Dynamic.Mission.THROTTLE],
                                 )

        if add_default_solver:
            prop_group.linear_solver = om.DirectSolver()
            prop_group.linear_solver.options["iprint"] = print_level

            prop_group.nonlinear_solver = om.NewtonSolver()
            prop_group.nonlinear_solver.options["err_on_non_converge"] = False
            prop_group.nonlinear_solver.options["solve_subsystems"] = True
            prop_group.nonlinear_solver.options["maxiter"] = 20
            prop_group.nonlinear_solver.options["iprint"] = print_level
            prop_group.nonlinear_solver.options["atol"] = atol
            prop_group.nonlinear_solver.options["rtol"] = rtol
            prop_group.nonlinear_solver.linesearch = om.BoundsEnforceLS()
            prop_group.linear_solver = om.DirectSolver(assemble_jac=True)

        if prop_group is not self:
            self.add_subsystem(
                'prop_group',
                prop_group,
                promotes=['*']
            )

    def add_flight_conditions(self, nn, input_speed_type=SpeedType.TAS):
        if input_speed_type is SpeedType.TAS:
            promotes_inputs = [("TAS", Dynamic.Mission.VELOCITY)]
            promotes_outputs = ["EAS", Dynamic.Mission.MACH]
        elif input_speed_type is SpeedType.EAS:
            promotes_inputs = ["EAS"]
            promotes_outputs = [("TAS", Dynamic.Mission.VELOCITY), Dynamic.Mission.MACH]
        elif input_speed_type is SpeedType.MACH:
            promotes_inputs = [Dynamic.Mission.MACH]
            promotes_outputs = ["EAS", ("TAS", Dynamic.Mission.VELOCITY)]

        self.add_subsystem(
            "flight_conditions",
            FlightConditions(num_nodes=nn, input_speed_type=input_speed_type),
            promotes_inputs=["rho", Dynamic.Mission.SPEED_OF_SOUND] + promotes_inputs,
            promotes_outputs=[Dynamic.Mission.DYNAMIC_PRESSURE] + promotes_outputs,
        )

    def add_excess_rate_comps(self, nn):
        self.add_subsystem(
            name='SPECIFIC_ENERGY_RATE_EXCESS',
            subsys=SpecificEnergyRate(num_nodes=nn),
            promotes_inputs=[Dynamic.Mission.VELOCITY, Dynamic.Mission.MASS,
                             (Dynamic.Mission.THRUST_TOTAL, Dynamic.Mission.THRUST_MAX_TOTAL),
                             Dynamic.Mission.DRAG],
            promotes_outputs=[(Dynamic.Mission.SPECIFIC_ENERGY_RATE,
                               Dynamic.Mission.SPECIFIC_ENERGY_RATE_EXCESS)]
        )

        self.add_subsystem(
            name='ALTITUDE_RATE_MAX',
            subsys=AltitudeRate(num_nodes=nn),
            promotes_inputs=[
                (Dynamic.Mission.SPECIFIC_ENERGY_RATE,
                 Dynamic.Mission.SPECIFIC_ENERGY_RATE_EXCESS),
                Dynamic.Mission.VELOCITY_RATE,
                Dynamic.Mission.VELOCITY],
            promotes_outputs=[
                (Dynamic.Mission.ALTITUDE_RATE,
                 Dynamic.Mission.ALTITUDE_RATE_MAX)])
